#*****************************************************************************
#
#    Sage Package "Modular Cohomology Rings of Finite Groups"
#
#    Copyright (C) 2009/2010
#         Simon A. King  <simon.king@nuigalway.ie> and
#         David J. Green <david.green@uni-jena.de>
#
#  Distributed under the terms of the GNU General Public License (GPL),
#  version 2 or later (at your choice)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************
r"""
Modular Cohomology Rings of Finite Groups

AUTHORS:

- Simon King  <simon.king@nuigalway.ie>
- David Green <david.green@uni-jena.de>

IMPLEMENTATION:

Our package is formed by Singular- and Gap-functions, as well as Python-, Cython- and C-code.

NOTE:

    Our package requires the `Small Groups
    <http://www-public.tu-bs.de:8080/~hubesche/small.html>`_ library
    of Hans Ulrich Besche, Bettina Eick and Eamonn O'Brien. If
    necessary, you can install it in Sage by
    ``install_package('database_gap')``.

The package "pGroupCohomology" comprises :func:`CohomologyRing`,
a constructor for the cohomology ring of a finite group `G` that
does caching etc. behind the scenes. The package also comprises eight
modules :mod:`~pGroupCohomology.factory`,
:mod:`~pGroupCohomology.cohomology`,
:mod:`~pGroupCohomology.modular_cohomology`,
:mod:`~pGroupCohomology.barcode`, :mod:`~pGroupCohomology.cochain`,
:mod:`~pGroupCohomology.resolution`, :mod:`~pGroupCohomology.mtx` and
:mod:`~pGroupCohomology.dickson` that do the real work.

In the first section of our introduction, we demonstrate the usage of
our package. After outlining the functionality, we go into more detail
and illustrate various ways of defining a cohomology ring, computing
the ring structure, and working in the cohomology ring.  Last, we
discuss how to save and load a cohomology ring on a local file system
and how to export and import it to a different platform.

In the second section, we give an overview on the theoretical
background of our package. The emphasis is on the construction of
parameters for the cohomology ring, which can then be used to prove
completeness of the ring structure by means of completeness criteria
that are due to different authors ([Benson]_, [GreenKing]_, [King]_,
[Symonds]_).  We illustrate this with step-by-step examples.

An overview of the various options that can be used is given in the
documentation of :func:`CohomologyRing`.

EXAMPLES:

1. Synopsis
===========

Our package is devoted to the cohomology rings of finite groups with
coefficients in `\mathbb F_p` (the finite field with `p` elements,
where `p` is a prime less than 255 that divides the group order).
Note that the first computation of the modular cohomology rings of all
groups of order 128 and the first computation of the mod-2 cohomology
ring of the third Conway group where obtained with our package; see
[GreenKing]_ and [KingGreenEllis]_ for more details. We computed the
mod-`p` cohomology rings for various interesting finite non prime
power groups and different primes `p`. Our results are available
`here <http://www.nuigalway.ie/maths/sk/Cohomology/rings/>`_. The non
prime power groups are not part of our web repository yet.

We expect the groups to be given either by their address in the Small
Groups Library, or as a group in Sage's Gap interface. In the latter
case and if `G` happens to be a prime power group, it is required that
the list of generators of `G` starts with a minimal generating set. If
this is not the case, a ``ValueError`` is raised. Internally, the group
will be transformed into a permutation group, which, in some cases, is
very difficult. So, it is recommended to start with a permutation group
right away.

The package is shipped with the cohomology rings for all groups of
order 64; the data are stored in folders and can simply be retrieved
by calling :func:`~pGroupCohomology.CohomologyRing` with appropriate
keys.  These data are provided to *all* users of the Sage installation,
thus we refer to it as the *public database*.

In addition, we provide the cohomology rings for all groups of order
128 and for all but six groups of order 243 in a web repository.
Unless the user requires a recomputation from scratch, these data will
be automatically downloaded. In future versions, this will also be
possible for various non prime power groups.

Any user can construct new databases; we refer to them as *private
database*. There is a default location for a private database,
namely in a sub-directory of ``DOT_SAGE``. But any location may be
chosen as well, provided that the user has write permission.

Cohomology rings are parent structures and are cached. That means, if
two groups are essentially equal then their cohomology rings will be
identic (and not only equal). Note for two groups being "essentially equal",
it does not suffice that they are isomorphic. All groups are supposed
to have a fixed list of generators. Two groups are "essentially equal",
if there exists a group isomorphism that sends the list of generators
of one group to an initial segment of the list of generators of the
other group. Moreover, two cohomology rings can only be identic if
they have the same location of data storage.

When a cohomology ring is created, the following locations are
searched, in order:

1. It is first tried to find the result in the cache.

2. It is tried to load it from the current private database.

3. If it is not there, the default is to try to find it in the public database.

4. Finally, by default it is attempted to download the result from a web
   repository.

The user has not necessarily write permission fo the public
database. Therefore, if data are found, symbolic links are created in
the private database that link to the known data in the public
database. If further computation requires more data, they are created
and stored in the private database.

If all this fails, a new ring is initialized in the current private
database. So, unless :meth:`~pGroupCohomology.CohomologyRing.public_db`
is used, the data for any cohomology ring will eventually be in the current
private database.

The data belonging to the cohomology ring of a prime power group (e.g., a minimal
free resolution, information on embedding of elementary abelian subgroups, standard
monomials etc.) are distributed in a folder tree, one tree for each group.

In the non prime power case, the situation is easier. It is not needed to
store small data pieces during the computation, and eventually the main
data are saved in a single file. Its default location is in the currently activated
private database. The computation does, however, rely on the cohomology rings
of various subgroups, e.g., a Sylow subgroup. So, again the data are distributed.

Creating a cohomology ring
--------------------------

In the following subsections, we explain here how to describe the
group whose modular cohomology ring shall be computed.

There are also various options that influence the algorithm and the use
of databases. For example, it might be instructive to use the protocol
option, providing the optional argument ``options='prot'`` or
``options=('prot',)``; the protocol gives detailed information on how
data are computed. If some timings shall be printed as well, one can
provide ``options=('prot','timing')``. Options can be changed at any
time using the method `~pGroupCohomology.cohomology.COHO.option`.

Usually, the package attempts to make use of existing data. However,
if the optional parameter ``from_scratch`` is set to ``True`` then
this cohomology ring is not loaded from the public database or the web
repository. Note, however, that this option only takes effect on the
ring itself, but not for cohomology rings of certain subgroups that
might be created during the computation.

Further options are described in the documentation of
:func:`~pGroupCohomology.CohomologyRing`.

... using the SmallGroups library
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

In our examples, we choose a temporary directory for our private data
base. Of course, in reality, a permanent directory would be chosen for
the databases. We do so in order to avoid a conflict with an existing
private database.
::

    sage: from pGroupCohomology import CohomologyRing
    sage: CohomologyRing.set_user_db(tmp_filename())
    sage: H0 = CohomologyRing(8,3)

Since the cohomology ring of ``SmallGroup(8,3)``, which is the
Dihedral Group of order 8, is in the public database shipped with our
package, it is simply loaded and already completely computed. It is
not needed to specify a prime to determine the coefficients, since the
group is of prime power order. Note that we also have a list of group
names and group descriptions -- hence, ``H0`` knows that it belongs to
the Dihedral Group of order 8::

    sage: H0
    H^*(D8; GF(2))
    sage: print H0
    Cohomology ring of Dihedral group of order 8 with coefficients in GF(2)
    <BLANKLINE>
    Computation complete
    Minimal list of generators:
    [c_2_2: 2-Cocycle in H^*(D8; GF(2)),
     b_1_0: 1-Cocycle in H^*(D8; GF(2)),
     b_1_1: 1-Cocycle in H^*(D8; GF(2))]
    Minimal list of algebraic relations:
    [b_1_0*b_1_1]

The generators are of type :class:`~pGroupCohomology.cochain.COCH`.
It is also possible to convert the cohomology ring into a quotient
ring in the Singular interface::

    sage: singular(H0)
    //   characteristic : 2
    //   number of vars : 3
    //        block   1 : ordering M
    //                  : names    c_2_2 b_1_0 b_1_1
    //                  : weights      2     1     1
    //                  : weights     -1     0     0
    //                  : weights      0    -1     0
    //        block   2 : ordering C
    // quotient ring from ideal
    _[1]=b_1_0*b_1_1

In Sage, parent structures should be unique. And indeed, cohomology
rings of *the same* group in *the same* database are unique::

    sage: H0 is CohomologyRing(8,3)
    True
    sage: H0 is loads(dumps(H0))
    True

We set up a different private database in a temporary directory, and
show that indeed two cohomology rings are only identical if their data
are stored in the same place. Moreover, we use an optional argument to
avoid that the ring is loaded from the public database or from a web
repository::

    sage: tmp = tmp_filename()
    sage: CohomologyRing.set_user_db(tmp)
    sage: H1 = CohomologyRing(8,3,from_scratch=True)
    sage: H1 is H0
    False

Since we used the option ``from_scratch=True``, it was not attempted
to download the cohomology ring from the web or load it from the
public database. Hence, beyond the basic setup, ``H1`` is unknown and
is thus not even equal to ``H0``::

    sage: H0 == H1
    False

We now compute ``H1`` (see below for the theoretical background) and
verify that then ``H0`` is equal (though not identical) to ``H1``::

    sage: H1.make()
    sage: H0 == H1
    True

Here is a similar example, using a non prime power group. In this
case, we need to provide the modulus by the optional parameter
``prime``.

::

    sage: HS6a = CohomologyRing(720,763,prime=2)
    sage: HS6a.make()
    sage: print HS6a
    <BLANKLINE>
    Cohomology ring of SmallGroup(720,763) with coefficients in GF(2)
    <BLANKLINE>
    Computation complete
    Minimal list of generators:
    [c_2_0: 2-Cocycle in H^*(SmallGroup(720,763); GF(2)),
     c_1_0: 1-Cocycle in H^*(SmallGroup(720,763); GF(2)),
     b_3_1: 3-Cocycle in H^*(SmallGroup(720,763); GF(2)),
     b_3_2: 3-Cocycle in H^*(SmallGroup(720,763); GF(2))]
    Minimal list of algebraic relations:
    [b_3_1*b_3_2]
    <BLANKLINE>
    sage: singular(HS6a)
    //   characteristic : 2
    //   number of vars : 4
    //        block   1 : ordering M
    //                  : names    c_2_0 c_1_0 b_3_1 b_3_2
    //                  : weights      2     1     3     3
    //                  : weights     -1    -1     0     0
    //                  : weights     -1     0     0     0
    //                  : weights      0     0    -1     0
    //        block   2 : ordering C
    // quotient ring from ideal
    _[1]=b_3_1*b_3_2

... using a group in the Gap interface
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Above, we mentioned that two cohomology rings are identic only if the
corresponding groups are essentially equal. Indeed, if we represent
the Dihedral Group of order 8 in a different way then we obtain a
different ring presentation. Note that, when referring to a group in
the Gap interface, we need to provide a name for that group, in one or
the other way. That name also determines the location of data storage,
and thus if two different names are provided then the corresponding
cohomology rings are not the same::

    sage: G = gap('DihedralGroup(8)')
    sage: H2 = CohomologyRing(G,GroupName='DihedralA')
    sage: G.SetName('"DihedralB"')
    sage: H3 = CohomologyRing(G)
    sage: H2 is H3
    False

Of course, when ``H2`` and ``H3`` are completely computed, they are
isomorphic to ``H1``, since they belong to isomorphic groups. However,
they do not have the same ring presentation. This is because the
presentation of the cohomology ring depends on the choice of a minimal
generating set of the group; and the minimal generating sets of
``gap('SmallGroup(8,3)')`` and of ``gap('DihedralGroup(8)')`` are
essentially different::

    sage: H2.make()
    sage: H3.make()
    sage: H2 == H3
    True
    sage: H2 == H1
    False

Indeed, the minimal relations of ``H1`` and ``H2`` look different::

    sage: print H1
    Cohomology ring of Dihedral group of order 8 with coefficients in GF(2)
    <BLANKLINE>
    Computation complete
    Minimal list of generators:
    [c_2_2: 2-Cocycle in H^*(D8; GF(2)),
     b_1_0: 1-Cocycle in H^*(D8; GF(2)),
     b_1_1: 1-Cocycle in H^*(D8; GF(2))]
    Minimal list of algebraic relations:
    [b_1_0*b_1_1]
    sage: print H2
    Cohomology ring of DihedralA with coefficients in GF(2)
    <BLANKLINE>
    Computation complete
    Minimal list of generators:
    [c_2_2: 2-Cocycle in H^*(DihedralA; GF(2)),
     b_1_0: 1-Cocycle in H^*(DihedralA; GF(2)),
     b_1_1: 1-Cocycle in H^*(DihedralA; GF(2))]
    Minimal list of algebraic relations:
    [b_1_1^2+b_1_0*b_1_1]

But of course, the rings are isomorphic.

Above, we computed the cohomology of some non prime power group. It is
in fact a symmetric group::

    sage: G = gap('SymmetricGroup(6)')
    sage: HS6b = CohomologyRing(G, prime=2, GroupName='SymmetricGroup(6)', from_scratch=True)
    sage: HS6b.make()
    sage: print HS6b
    <BLANKLINE>
    Cohomology ring of SymmetricGroup(6) with coefficients in GF(2)
    <BLANKLINE>
    Computation complete
    Minimal list of generators:
    [c_2_1: 2-Cocycle in H^*(SymmetricGroup(6); GF(2)),
     c_1_0: 1-Cocycle in H^*(SymmetricGroup(6); GF(2)),
     b_3_2: 3-Cocycle in H^*(SymmetricGroup(6); GF(2)),
     b_3_3: 3-Cocycle in H^*(SymmetricGroup(6); GF(2))]
    Minimal list of algebraic relations:
    [b_3_2*b_3_3]
    <BLANKLINE>

Obviously the two ring presentations for the mod-2 cohomology of the
symmetric group of rank 6 are different. This is since the ring
presentation will essentially depend on the choice of generators of
the Sylow subgroup (which here is chosen by the program).


Computations in a cohomology ring
---------------------------------

Ring theoretic invariants
^^^^^^^^^^^^^^^^^^^^^^^^^

Up to now, we implemented the computation of four ring theoretic
invariants, namely the dimension, the depth, the a-invariants and
the Poincar\'e series::

    sage: H2.dimension()
    2
    sage: H2.depth()
    2
    sage: H2.a_invariants()
    [-Infinity, -Infinity, -2]
    sage: H2.poincare_series()
    1/(t^2 - 2*t + 1)

Of course, they are the same for both presentations of the cohomology
ring of the dihedral group of order 8::

    sage: H1.dimension()
    2
    sage: H1.depth()
    2
    sage: H1.a_invariants()
    [-Infinity, -Infinity, -2]
    sage: H1.poincare_series()
    1/(t^2 - 2*t + 1)

Actually we implemented two essentially different ways of computating
the Poincar\'e series. The method
:meth:`~pGroupCohomology.cohomology.COHO.poincare_without_parameters`
is usually much slower than
:meth:`~pGroupCohomology.cohomology.COHO.poincare_series`, but of
course the results are the same::

    sage: H2.poincare_without_parameters()
    1/(t^2 - 2*t + 1)

After computing the ring structure, generators and relations of the
cohomology ring are present. Generator number zero is the
multiplicative unit of the underlying field. The other generators are
represented by cochains (:class:`~pGroupCohomology.cochain.COCH`, or
:class:`~pGroupCohomology.cochain.MODCOCH` for non prime power groups)::

    sage: H1.gens()
    [1, c_2_2: 2-Cocycle in H^*(D8; GF(2)), b_1_0: 1-Cocycle in H^*(D8; GF(2)), b_1_1: 1-Cocycle in H^*(D8; GF(2))]

The relations are given as strings::

    sage: H1.rels()
    ['b_1_0*b_1_1']

The relations can be interpreted as elements of the cohomology ring,
and indeed they vanish::

    sage: print H1(H1.rels()[0])
    2-Cocycle in H^*(D8; GF(2)),
    represented by
    [0 0 0]

We can also compute the nil radical, i.e., the ideal formed by all
nilpotent elements. It is given in the Singular interface, as an ideal
in a free graded commututive ring. In particular, all elements of the
relation ideal of the cohomology ring are in the nil radical. For the
dihedral group of order 8, the nil radical vanishes::

    sage: H1.nil_radical()
    0
    sage: H2.nil_radical()
    0

Here are the data for the mod-2 cohomology of the symmetric group of
degree 6::

    sage: HS6a.dimension()
    3
    sage: HS6a.depth()
    3
    sage: HS6a.a_invariants()
    [-Infinity, -Infinity, -Infinity, -3]
    sage: HS6a.poincare_series()
    (-t^2 + t - 1)/(t^5 - 2*t^4 + t^3 - t^2 + 2*t - 1)
    sage: HS6a.poincare_without_parameters()
    (-t^2 + t - 1)/(t^5 - 2*t^4 + t^3 - t^2 + 2*t - 1)
    sage: HS6a.nil_radical()
    0


Induced maps
^^^^^^^^^^^^

As we have seen before, the presentation of the cohomology ring
depends on the choice of a minimal generating set of the underlying
finite `p`-group (resp. Sylow `p`-subgroup). For our purposes, we always
assume that a finite group has a fixed set of generators, and we call
two groups *equivalent*, if there is a group isomorphism that maps the
given list of generators of one group to an initial segment of the list
of generators of the other group.

:meth:`~pGroupCohomology.cohomology.COHO.group` returns a permutation
group in the gap interface equivalent to the one that was used to set up the
cohomology ring::

    sage: G1 = H1.group()
    sage: G1
    Group( [ (1,2)(3,8)(4,6)(5,7), (1,3)(2,5)(4,7)(6,8) ] )
    sage: G2 = H2.group()
    sage: G2
    Group( [ (1,2)(3,8)(4,6)(5,7), (1,3,4,7)(2,5,6,8), (1,4)(2,6)(3,7)(5,8) ] )

Apparently, ``G1`` and ``G2`` are different permutation groups. But in
fact they are isomorphic. Note that we could get a group isomorphism
by Gap's function ``IsomorphismGroups``. But the result is not
unique. In order to have reproducible doc tests, we provide an
explicit isomorphism::

    sage: phi = gap('GroupHomomorphismByImages(Group([(1,2)(3,8)(4,6)(5,7),(1,3)(2,5)(4,7)(6,8)]), Group([(1,2)(3,8)(4,6)(5,7),(1,3,4,7)(2,5,6,8),(1,4)(2,6)(3,7)(5,8) ]), [(1,2)(3,8)(4,6)(5,7), (1,3)(2,5)(4,7)(6,8)], [(1,6)(2,4)(3,5)(7,8), (1,8)(2,7)(3,6)(4,5)])')
    sage: phi.IsInjective()
    true
    sage: phi.IsSurjective()
    true

Of course, the group isomorphism induces an isomorphism of the
corresponding cohomology rings. After ensuring that ``phi`` is printed
nicely, we obtain the induced map::

    sage: phi.SetName('"phi"')
    sage: phi_star = H2.hom(phi,H1)
    sage: phi_star
    phi^*

Indeed, we can verify that the image of the relation of ``H2`` is zero
in ``H1``::

    sage: H2.rels()
    ['b_1_1^2+b_1_0*b_1_1']
    sage: phi_star(H2('b_1_1'))^2 + phi_star(H2('b_1_0'))*phi_star(H2('b_1_1'))
    (phi^*(b_1_1))**2+(phi^*(b_1_0))*(phi^*(b_1_1)): 2-Cocycle in H^*(D8; GF(2))
    sage: print phi_star(H2('b_1_1'))^2 + phi_star(H2('b_1_0'))*phi_star(H2('b_1_1'))
    2-Cocycle in H^*(D8; GF(2)),
    represented by
    [0 0 0]

We can also compute the inverse of ``phi_star``, compose the two maps,
and see that it is the identity map::

    sage: phi_star_inv = phi_star^(-1)
    sage: phi_star_inv
    (phi^*)^(-1)
    sage: [(phi_star_inv*phi_star)(X) == X for X in H2.gens()]
    [True, True, True, True]
    sage: [(phi_star*phi_star_inv)(X) == X for X in H1.gens()]
    [True, True, True, True]

It is possible to convert an induced homomorphism into a map of
quotient rings in the Singular interface. This allows for working with
cohomology ring elements of very high degrees. However, it is always
needed to take care of the ``basering`` in Singular::

    sage: S_phi_star = singular(phi_star)
    sage: singular(H2).set_ring()
    sage: I = singular.ideal(['c_2_2^50','b_1_0^50','b_1_1^50'])
    sage: singular(H1).set_ring()
    sage: S_phi_star(I)
    b_1_1^100+c_2_2^2*b_1_1^96+c_2_2^16*b_1_1^68+c_2_2^18*b_1_1^64+c_2_2^32*b_1_1^36+c_2_2^34*b_1_1^32+c_2_2^48*b_1_1^4+c_2_2^50,
    b_1_1^50+b_1_0^2*b_1_1^48+b_1_0^16*b_1_1^34+b_1_0^18*b_1_1^32+b_1_0^32*b_1_1^18+b_1_0^34*b_1_1^16+b_1_0^48*b_1_1^2+b_1_0^50,
    b_1_1^50
        
Note that Singular does not do automatic reduction in quotient
rings. So, eventually we do the reductions explicitly, in two ways,
with the same result::

    sage: S_phi_star(I).reduce(singular('ideal(0)'))
    b_1_1^100+c_2_2^2*b_1_1^96+c_2_2^16*b_1_1^68+c_2_2^18*b_1_1^64+c_2_2^32*b_1_1^36+c_2_2^34*b_1_1^32+c_2_2^48*b_1_1^4+c_2_2^50,
    b_1_1^50+b_1_0^50,
    b_1_1^50
    sage: singular(H2).set_ring()
    sage: I = I.reduce(singular('ideal(0)'))
    sage: singular(H1).set_ring()
    sage: S_phi_star(I)
    b_1_1^100+c_2_2^2*b_1_1^96+c_2_2^16*b_1_1^68+c_2_2^18*b_1_1^64+c_2_2^32*b_1_1^36+c_2_2^34*b_1_1^32+c_2_2^48*b_1_1^4+c_2_2^50,
    b_1_1^50+b_1_0^2*b_1_1^48+b_1_0^16*b_1_1^34+b_1_0^18*b_1_1^32+b_1_0^32*b_1_1^18+b_1_0^34*b_1_1^16+b_1_0^48*b_1_1^2+b_1_0^50,
    b_1_1^50+b_1_0*b_1_1^49+b_1_0^16*b_1_1^34+b_1_0^17*b_1_1^33+b_1_0^32*b_1_1^18+b_1_0^33*b_1_1^17+b_1_0^48*b_1_1^2+b_1_0^49*b_1_1
    sage: S_phi_star(I).reduce(singular('ideal(0)'))
    b_1_1^100+c_2_2^2*b_1_1^96+c_2_2^16*b_1_1^68+c_2_2^18*b_1_1^64+c_2_2^32*b_1_1^36+c_2_2^34*b_1_1^32+c_2_2^48*b_1_1^4+c_2_2^50,
    b_1_1^50+b_1_0^50,
    b_1_1^50

For a non prime power group, a permutation representation equivalent to the given group,
a subgroup which was used for the stable element method and a Sylow `p`-subgroup are available
as well::

    sage: HS6a.group()
    Group( [ (1,2), (1,2,3,4,5,6) ] )
    sage: HS6a.subgroup()
    Group( [ (1,4)(2,6), (1,3)(4,5), (2,6), (1,4)(3,5) ] )
    sage: HS6a.sylow_subgroup()
    Group( [ (1,4)(2,6), (1,3)(4,5), (2,6) ] )

The dihedral group of order 8 is a subgroup of the symmetric group of
rank 6. We compute the ring homomorphism that is induced by the
embedding. In order to get a reproducible result, we define the
group homomorphism explicitly::

    sage: phi = gap('GroupHomomorphismByImages( Group( [ (1,2)(3,8)(4,6)(5,7), (1,3)(2,5)(4,7)(6,8) ] ), Group( [ (1,2), (1,2,3,4,5,6) ] ), [ (1,2)(3,8)(4,6)(5,7), (1,3)(2,5)(4,7)(6,8) ], [ (1,3), (1,2)(3,4) ] )')
    sage: phi_star = HS6a.hom(phi,H0)
    sage: [H0.element_as_polynomial(phi_star(x)) for x in HS6a.gens()]   
    [1: 0-Cocycle in H^*(D8; GF(2)), b_1_1^2+c_2_2: 2-Cocycle in H^*(D8; GF(2)), b_1_0: 1-Cocycle in H^*(D8; GF(2)), c_2_2*b_1_0: 3-Cocycle in H^*(D8; GF(2)), c_2_2*b_1_1: 3-Cocycle in H^*(D8; GF(2))]
    sage: singular(H0).set_ring()
    sage: singular(phi_star)
    s...[1]=b_1_1^2+c_2_2
    s...[2]=b_1_0
    s...[3]=c_2_2*b_1_0
    s...[4]=c_2_2*b_1_1

Here is the kernel of the induced map::

    sage: phi_star.preimage()
    b_3_1+c_2_0*c_1_0,
    c_1_0*b_3_2
    sage: phi_star(HS6a('b_3_1+c_2_0*c_1_0')).as_polynomial()
    '0'
    sage: phi_star(HS6a('b_3_2*c_1_0')).as_polynomial()
    '0'

Massey products
^^^^^^^^^^^^^^^

We can compute (set valued) higher Massey products and the (single
valued) restricted Massey powers introduced by [Kraines]_ for cocycles
of prime power groups. For non prime power groups, we have so far only
an experimental implementation for restricted Massey powers.

We first study some examples on ``H0`` (cohomology of the Dihedral Group
of order 8) that we have defined above. Note that we are transforming
the returned sets into sorted lists, in order
to have reproducible doc tests::

    sage: sorted(list(H0.massey_products(H0.2,H0.3,H0.2)), cmp=lambda X,Y:cmp(X.name(),Y.name()))
    [0: 2-Cocycle in H^*(D8; GF(2)), b_1_0^2: 2-Cocycle in H^*(D8; GF(2))]
    sage: sorted(list(H0.massey_products(H0.2,H0.3,H0.2,H0.3)), cmp=lambda X,Y:cmp(X.name(),Y.name()))
    [b_1_0^2+c_2_2: 2-Cocycle in H^*(D8; GF(2)),
     b_1_1^2+b_1_0^2+c_2_2: 2-Cocycle in H^*(D8; GF(2)),
     b_1_1^2+c_2_2: 2-Cocycle in H^*(D8; GF(2)),
     c_2_2: 2-Cocycle in H^*(D8; GF(2))]
    sage: sorted(list(H0.massey_products(H0.2*H0.1,H0.3*H0.1,H0.2*H0.1)), cmp=lambda X,Y:cmp(X.name(),Y.name())) # long time
    [0: 8-Cocycle in H^*(D8; GF(2)),
     c_2_2*b_1_0^6: 8-Cocycle in H^*(D8; GF(2)),
     c_2_2*b_1_0^6+c_2_2^2*b_1_0^4: 8-Cocycle in H^*(D8; GF(2)),
     c_2_2*b_1_0^6+c_2_2^2*b_1_0^4+c_2_2^3*b_1_0^2: 8-Cocycle in H^*(D8; GF(2)),
     c_2_2*b_1_0^6+c_2_2^3*b_1_0^2: 8-Cocycle in H^*(D8; GF(2)),
     c_2_2^2*b_1_0^4: 8-Cocycle in H^*(D8; GF(2)),
     c_2_2^2*b_1_0^4+c_2_2^3*b_1_0^2: 8-Cocycle in H^*(D8; GF(2)),
     c_2_2^3*b_1_0^2: 8-Cocycle in H^*(D8; GF(2))]

If one is interested in just a single element of the Massey products,
one can use the option ``all=False``::

    sage: H0.massey_products(H0.2*H0.1,H0.3*H0.1,H0.2*H0.1,H0.3*H0.1, all=False)
    set([c_2_2^5: 10-Cocycle in H^*(D8; GF(2))])

By the `i`-th restricted Massey power of a cocycle `C` in a cohomology
ring with coefficients `\mathbb F_p`, we mean the restricted
`p^i`-fold Massey product of `C` as introduced by D. Kraines. If it is
defined, then it is a single cocycle::

    sage: H = CohomologyRing(27,3)
    sage: H.make()
    sage: H.8
    a_3_4: 3-Cocycle in H^*(E27; GF(3))
    sage: H.8.massey_power()
    <a_3_4; 1>: 8-Cocycle in H^*(E27; GF(3))
    sage: H.element_as_polynomial(_)
    b_2_0^2*a_1_1*a_3_5+b_2_0^2*a_1_0*a_3_4+b_2_0*c_6_8: 8-Cocycle in H^*(E27; GF(3))

It is known that in odd degree and for `p>2` it can be expressed as
minus the composition of the first Steenrod power with the Bockstein
operator. This allows for a verification of the above result, see
:meth:`~pGroupCohomology.cochain.COCH.massey_power` for details.

Massey products allow to distinguish certain isomorphic cohomology
rings, so, they contain information that go beyond the ring
structure. An easy example is provided by the cohomology rings of the
cyclic groups of order 3 and order 9. The rings are isomorphic.
::

    sage: HC3 = CohomologyRing(3,1)
    sage: HC3.make()
    sage: HC9 = CohomologyRing(9,1)
    sage: HC9.make()
    sage: print HC3
    Cohomology ring of Small Group number 1 of order 3 with coefficients in GF(3)
    <BLANKLINE>
    Computation complete
    Minimal list of generators:
    [c_2_0: 2-Cocycle in H^*(SmallGroup(3,1); GF(3)),
     a_1_0: 1-Cocycle in H^*(SmallGroup(3,1); GF(3))]
    Minimal list of algebraic relations:
    []
    sage: print HC9
    Cohomology ring of Small Group number 1 of order 9 with coefficients in GF(3)
    <BLANKLINE>
    Computation complete
    Minimal list of generators:
    [c_2_0: 2-Cocycle in H^*(SmallGroup(9,1); GF(3)),
     a_1_0: 1-Cocycle in H^*(SmallGroup(9,1); GF(3))]
    Minimal list of algebraic relations:
    []

Note that the relation ``a_1_0^2`` is not listed, since the relation
`x^2=0` holds for any cocycle `x` of odd degree, for `p>2`, and is
thus somehow redundant. The 1st restricted Massey power of the degree
one generator is non-trivial for the cyclic group of order 3, but only
the 2nd restricted Massey power of the degree one generator of the
cohomology ring of the cyclic group of order 9 is non-trivial::

    sage: HC3.element_as_polynomial(HC3.2.massey_power())
    -c_2_0: 2-Cocycle in H^*(SmallGroup(3,1); GF(3))
    sage: HC9.element_as_polynomial(HC9.2.massey_power())
    0: 2-Cocycle in H^*(SmallGroup(9,1); GF(3))
    sage: HC9.element_as_polynomial(HC9.2.massey_power(2))
    -c_2_0: 2-Cocycle in H^*(SmallGroup(9,1); GF(3))

Bar Codes
^^^^^^^^^

Cohomology rings contain much information on a group. Unfortunately,
it is, in general, quite difficult to test whether two graded
commutative rings are isomorphic. Hence, it is of interest to consider
invariants that can be more easily compared, such as the Poincar\'e
series or the depth. Bar codes are not ring theoretic invariants,
but are defined in terms of the finite group whose cohomology is
being studied.

Any sequence of group homomorphisms `\phi_i: G_i \to G_{i+1}` with
`i=1,...,n` gives rise to a series of induced homomorphisms of
cohomology rings. Note that we insist that the sequence starts with
the group and ends with the trivial group. It is easy to get such
sequences in the case of prime power groups (e.g., upper central
series, derived series, `p`-central series). For non prime power
groups, this may be more difficult.

Persistent group cohomology, introduced in [EllisKing]_, asks how long
cocycles in a given degree "survive" being mapped by the induced
homomorphisms. In our paper, we observe a certain correspondence to
periodicity of coclass trees, and suggest that persistent group
homology may be an interesting theoretical tool. In addition, it can
be seen as an invariant of groups.

For a given degree `d`, the persistent cohomology can be described by
an upper triangular matrix of non-negative integers, where entry `i,j`
(`i\le j`) is the dimension of the image of the degree `d` part of
`H^*(G_j)` in `H^*(G_i)` under the induced homomorphism given by the
composition of `\phi_i,\phi_{i+1},...,\phi_{j-1}`, including the
case `i=j` in which the matrix entry simply gives the dimension of the
degree `d` part of `H^*(G_i)`.

As usual, the sequence of dimensions sorted by degree gives rise to a
Poincar\'e series. Hence, an upper triangular matrix of rational
functions results and this provides information for all degrees.

We like to emphasize that the persistence matrices are very easy to
compare and provide a quite strong tool for distinguishing groups.
While the cohomology ring only depends on the group algebra, the bar
codes may contain additional information on the group itself, by
taking into account the normal series. However, the following example
shows that bar codes contain information that even goes beyond the
isomorphism types of the normal subgroups and quotient groups.
::

    sage: H158 = CohomologyRing(64,158)
    sage: H160 = CohomologyRing(64,160)

These cohomology rings are part of our database, hence, they are
already completely known. The Poincar\'e series, the a-invariants, the
degrees of generators and of relations of the cohomology rings
coincide::

    sage: H158.poincare_series()
    (t^4 + t^3 + t^2 + t + 1)/(t^6 - 2*t^5 + 3*t^4 - 4*t^3 + 3*t^2 - 2*t + 1)
    sage: H160.poincare_series()
    (t^4 + t^3 + t^2 + t + 1)/(t^6 - 2*t^5 + 3*t^4 - 4*t^3 + 3*t^2 - 2*t + 1)
    sage: H158.degvec
    [4, 4, 1, 1, 1, 3, 3]
    sage: H160.degvec
    [4, 4, 1, 1, 1, 3, 3]
    sage: H158.set_ring()
    sage: [singular('deg(%s)'%r) for r in H158.rels()]
    [2, 2, 3, 3, 4, 4, 5, 6, 6, 6]
    sage: H160.set_ring()
    sage: [singular('deg(%s)'%r) for r in H160.rels()]
    [2, 2, 3, 3, 4, 4, 5, 6, 6, 6]
    sage: H158.a_invariants()
    [-Infinity, -Infinity, -2]
    sage: H160.a_invariants()
    [-Infinity, -Infinity, -2]

We consider here the bar codes associated with the upper central
series. It turns out that the non-trivial terms of the upper central
series and the resulting factor groups are isomorphic::

    sage: G158 = gap('SmallGroup(64,158)')
    sage: G160 = gap('SmallGroup(64,160)')
    sage: [(G.IdGroup(), (G158/G).IdGroup()) for G in G158.UpperCentralSeries()]
    [([ 64, 158 ], [ 1, 1 ]),
     ([ 16, 2 ], [ 4, 2 ]),
     ([ 4, 2 ], [ 16, 11 ]),
     ([ 1, 1 ], [ 64, 158 ])]
    sage: [(G.IdGroup(), (G160/G).IdGroup()) for G in G160.UpperCentralSeries()]
    [([ 64, 160 ], [ 1, 1 ]),
     ([ 16, 2 ], [ 4, 2 ]),
     ([ 4, 2 ], [ 16, 11 ]),
     ([ 1, 1 ], [ 64, 160 ])]

Nevertheless, the groups can be distinguished using the bar codes
associated with the upper central series::

    sage: B158 = H158.bar_code('UpperCentralSeries')
    sage: B160 = H160.bar_code('UpperCentralSeries')
    sage: B158 == B160
    False

Indeed, the bar codes differ in degree 3; graphically::

    sage: print B158[3]
            *
            *
          *-*
          *-*
          *
          *
          *
          *
          *
          *
        *-*
        *-*
        *
        *
      *
      *
      *
      *
    *
    *
    *
    *
    sage: print B160[3]
            *
            *
          *-*
          *
          *
          *
          *
          *
          *
          *
        *-*-*
        *-*
        *
      *-*
      *
      *
      *
    *
    *
    *
    *

These pictures (bar codes) can be interpreted as follows. Let `G` be a
finite `p`-group and let `G=G_0 > G_1 > ... > G_n > 1` be a normal
series; in our example, we have `n=2`. The first `n+1` columns of the
bar code correspond to the normal subgroups groups `G_n, G_{n-1},...,
G_0`, while the last `n` columns correspond to the factor groups
`G/G_n, G/G_{n-1},..., G/G_1`. We consider the sequence of group
homomorphisms given by inclusions and quotients. The stars in the bar
code correspond to basis vectors of the degree `d` part of the
cohomology rings of the respective groups. Stars are connected by a
line (i.e., a hyphen) if the corresponding basis vectors are mapped to
each other by the induced maps (which of course go from right to
left).

In terms of the persistence matrix::

    sage: B158[3].matrix()
    [ 4  0  0  0  0]
    [ 0  4  0  0  0]
    [ 0  0  4  2  0]
    [ 0  0  0 10  2]
    [ 0  0  0  0  4]
    sage: B160[3].matrix()
    [ 4  0  0  0  0]
    [ 0  4  1  0  0]
    [ 0  0  4  2  1]
    [ 0  0  0 10  2]
    [ 0  0  0  0  4]

The bar codes in degree 3 can also be computed directly, and of course
the result coincides with the bar codes obtained from the "full" bar
codes (based on Poincar\'e series)::

    sage: B158[3] == H158.bar_code('UpperCentralSeries',degree=3)
    True
    sage: B160[3] == H160.bar_code('UpperCentralSeries',degree=3)
    True

Since apparently the ``(2,4)`` entries of the persistence matrices
differ, we show here how the Poincar\'e series in that position look
like::

    sage: B158.matrix()[2,4]
    2*t^2 + 2*t + 1
    sage: B160.matrix()[2,4]
    t^3 + 2*t^2 + 2*t + 1

Hence, the iamge of the map `H^*(G/G_2) \to H^*(G)` induced by the
quotient map contains only finitely many cocycles, where `G` is group
number 158 or 160 of order 64, and `G_2` is the second term of the
upper central series.

Essential and Depth Essential Ideals
------------------------------------

The essential ideal of a cohomology ring is formed by all those elements
whose restrictions to all subgroups vanish. The depth essential ideal
at rank `r` is formed by all those elements that vanish on the centralisers
of all `p`-elementary abelian subgroups of rank `r`. These ideals can be
computed using :meth:`~pGroupCohomology.cohomology.COHO.essential_ideal`
and :meth:`~pGroupCohomology.cohomology.COHO.depth_essential_ideal`.

If `r` is at most the depth of the cohomology ring, then the depth essential
ideal vanishes. It is conjectured by J. Carlson that it is non-zero whenever
`r` exceeds the depth.
::

    sage: D = CohomologyRing(8,3)
    sage: D.make()
    sage: Q = CohomologyRing(8,4)
    sage: Q.make()
    sage: SD = CohomologyRing(16,8)
    sage: SD.make()
    sage: S6 = CohomologyRing(720,763,prime=2)
    sage: S6.make()

The quaternion group is the smallest group that has a non-vanishing
essential ideal::

    sage: Q.essential_ideal()
    a_1_0^2,
    a_1_0*a_1_1

The dihedral group and the group of order 720 (which is
the symmetric group on 6 elements) have no essential
classes, for different reasons::

    sage: S6.option('prot')
    sage: S6.essential_ideal()
    The group is not of prime power order -- no essential ideal
    0
    sage: D.essential_ideal()
    The depth exceeds the Duflog bound -- no essential ideal
    0
    sage: D.option('noprot')

It used to be a conjecture that the product of any two essential
classes vanishes. But then it was found that the cohomology of the
Sylow 2-subgroup of `U_3(4)`, which is of order 64, has exactly
one essential class that decomposes into essential classes of smaller
degrees. We verify this property::

    sage: H = CohomologyRing(64,245)
    sage: H
    H^*(Syl2(U3(4)); GF(2))
    sage: I = H.essential_ideal()    #long time
    sage: singular(H).set_ring()
    sage: (I*I).NF('std(0)').interred()  #long time
    a_4_8*a_6_8*a_1_0^3*a_1_3
    sage: singular('NF((a_4_8*a_6_8*a_1_0^3*a_1_3)^2, std(0))')
    0

Until recently, that was the only known counterexample to the
conjecture. However, using this package, it was found that the
cohomology ring of the direct product of the cyclic group of order two
and the Sylow 2-subgroup of `U_3(4)` has the same property.

Last, we illustrate Carlson's depth essential conjecture with some
group of order 64::

    sage: H = CohomologyRing(64,23)
    sage: H.CenterRk
    2
    sage: H.depth()
    3
    sage: H.dimension()
    4
    sage: H.depth_essential_ideal(2)
    0
    sage: H.depth_essential_ideal(3)
    0
    sage: H.depth_essential_ideal(4)
    a_1_0,
    a_1_1,
    a_2_0

Data storage and data relocation
--------------------------------

Modular cohomology rings of finite groups can be very complicated.
When computing the ring structure, many data pieces are
needed --- but not all pieces at the same time. Therefore we save the
various pieces in separate files, and reload them when they are
needed. In some examples, the memory and time consumption of saving
*all* data in a single file would not be reasonable. Hence, 'saving' a
cohomology ring to some file ``f`` means that ``f`` only contains the
information needed to find the files in which the various data pieces
are stored.

This method is fast, but has a clear disadvantage: If you wish to
transfer your cohomology computations to a different computer, it does
not suffice to copy one file for one cohomology ring. *All* data
pieces must be copied. Moreover, when the data are in a different
location, then the information stored in ``f`` would point to
non-existing paths. We tried to address these problems in the least
inconvenient way, as explained in the following sub-sections.

Data storage
^^^^^^^^^^^^

We first describe the situation for prime power groups. We do some new
computations from scratch, using a temporary directory.
::
  
    sage: from pGroupCohomology import CohomologyRing
    sage: tmp = tmp_filename()
    sage: CohomologyRing.set_user_db(tmp)
    sage: H = CohomologyRing(8,3,from_scratch=True)
    sage: H.make()

By default, when computing a cohomology ring ``H``, the result is
automatically saved in a file ``H.autosave_name()``. Cohomology rings
are unique parent structures, and therefore reloading from that file
only creates a new reference to ``H``::

    sage: K = load(H.autosave_name())
    sage: K is H
    True
    sage: loads(dumps(H)) is H
    True

By the command ``CohomologyRing.set_user_db(tmp)``, all data in the
user database are stored in sub-folders of the folder ``tmp``. The
names of these folders are composed as follows.

1. The basic folder ``tmp`` is known to ``H`` by the attribute
   ``root``. :func:`~pGroupCohomology.CohomologyRing` normalizes
   the given folder name ``tmp`` so that it does not contain double
   slashes. In order to not break the following doc test, we remove
   double slashes from ``tmp``.
   ::

       sage: tmp = tmp.replace('//','/')
       sage: H.root == tmp+'/'
       True

2. Any cohomology ring has been assigned a "stem name" that determines
   the names of sub-folders and file names associated with the ring.
   If the cohomology ring belongs to a group in the Small Groups Library,
   its stem name is formed using the address in the Small Groups Library.
   Otherwise, it is obtained from the group name used in the definition
   of the cohomology ring::

       sage: H = CohomologyRing(8,3)
       sage: G = gap('DihedralGroup(8)')
       sage: K = CohomologyRing(G, GroupName = 'DihedralA')
       sage: G.SetName('"DihedralB"')
       sage: L = CohomologyRing(G)
       sage: H.GStem
       '8gp3'
       sage: K.GStem
       'DihedralA'
       sage: L.GStem
       'DihedralB'

3. Data belonging to a cohomology ring can be found in a folder that is 
   stored as the attribute "gps_folder", and is composed by the root and
   the stem name::

       sage: H.gps_folder == H.root+H.GStem+'/'
       True

The situation for non prime power groups is different. Here, the data
structure for the ring itself is somewhat easier, so, main data are
saved in just one file.  On the other hand, the computations rely on
the stable element method and thus involve the cohomology rings of
various subgroups. When reloading, these are searched in the cache or
in the public or private databases or in the web repository, and are
re-computed if necessary. Here, we show that the rings are cached::

    sage: HS6a = CohomologyRing(720,763,prime=2)
    sage: HS6a.make()
    sage: HS6a is loads(dumps(HS6a))
    True
    sage: save(HS6a, HS6a.autosave_name())
    sage: HS6a is load(HS6a.autosave_name())
    True

Data relocation
^^^^^^^^^^^^^^^

If you wish to transfer data of a cohomology ring ``H`` to a different
location, you may do so. But please make sure that (in the case of
prime power groups) ``H.gps_folder`` keeps its name (only the parent
directory is allowed to change!), and that all files in
``H.gps_folder`` and its sub-folders are preserved. Then, by
``CohomologyRing.set_user_db``, one can activate the private database
in the new location, and reload the ring.  In order to demonstrate
that it really is the same ring, we provide ``H`` with an additional
attribute before saving and moving the data::

    sage: H.setprop('info', 'The cohomology data were moved')
    sage: save(H,H.autosave_name())
    sage: tmp2 = tmp_filename()
    sage: import os
    sage: os.rename(tmp,tmp2)
    sage: CohomologyRing.set_user_db(tmp2)
    sage: M = CohomologyRing(8,3)
    sage: M is H
    False
    sage: print M
    Cohomology ring of Dihedral group of order 8 with coefficients in GF(2)
    <BLANKLINE>
    Computation complete
    Minimal list of generators:
    [c_2_2: 2-Cocycle in H^*(D8; GF(2)),
     b_1_0: 1-Cocycle in H^*(D8; GF(2)),
     b_1_1: 1-Cocycle in H^*(D8; GF(2))]
    Minimal list of algebraic relations:
    [b_1_0*b_1_1]
    sage: M.info
    'The cohomology data were moved'

2. Background for the Cohomology Computation of Prime Power Groups
==================================================================

Elementary abelian subgroups and choice of generators
-----------------------------------------------------

Let `G` be a non-abelian finite `p`-group that is not a generalised
quaternion group. Let `S_1` be its greatest central elementary abelian
subgroup, and let `S_2,...,S_n` be representatives for the conjugacy
classes of maximal elementary abelian subgroups of `G`.

Our approach relies on knowing these subgroups, to which we refer as
*special subgroups*. When defining a cohomology ring, the special
subgroups are computed using Gap functions written by David Green.

In our example, `G` is the dihedral group of order 8, which is the
third group of order 8 in the Small Groups Library. We set up the
cohomology ring in a temporary directory. The special subgroups can be
retrieved as follows::

    sage: from pGroupCohomology import CohomologyRing
    sage: tmp = tmp_filename()
    sage: CohomologyRing.set_user_db(tmp)
    sage: H = CohomologyRing(8,3, from_scratch=True)
    sage: sorted(H.subgroups().items())
    [((2, 1), H^*(SmallGroup(2,1); GF(2))), ((4, 2), H^*(SmallGroup(4,2); GF(2)))]

Hence, ``H.subgroups()`` is a dictionary, whose keys correspond to
elementary abelian groups in the SmallGroups library, and whose values
are the corresponding cohomology rings.

[Green]_ provides an algorithm for the computation of minimal
projective resolutions of finite `p`-groups, using non-commutative
Groebner bases. He implemented it in C-programs, and we provide a
Cython wrapper in :mod:`pGroupCohomology.resolution`.

Knowing a resolution out to degree `d` allows for computing a minimal
set of cohomology generators out to degree `d` and a minimal set of
algebraic relations of degree up to `d` for these generators. It pays
off to choose the generators in a special way, as follows:

1. A cohomology class of `G` is nilpotent if and only if its 
   restrictions to `S_2,...,S_n` are nilpotent. Since the cohomology
   rings of elementary abelian groups are either polynomial rings
   or the tensor product of a polynomial ring with an outer algebra
   (if `p` is odd), it is easy to test whether a cohomology class of `G`
   is nilpotent, even if the whole cohomology ring is not known yet.
   We choose as many as possible new nilpotent generators in degree `d`.
2. After step 1, continue with choosing as many as possible new
   generators in degree `d` with nilpotent restriction to `S_1`.
3. After step 2, include the remaining new generators in degree 
   `d`. These are Duflot regular.

If `z` is the rank of the centre of `G` then there are exactly `z`
Duflot regular generators, which was proved by [Kuhn]_.  David Green
introduced a monomial ordering on the cohomology ring of `G` that, in
the first place, depends on the degree of a monomial, but also depends
on the number of nilpotent respectively Duflot regular generators
occuring in the monomial. It turns out that this monomial ordering
magically simplifies many computations. We now compute the cohomology
ring ``H`` out to degree 2, and show the generators and relations.
::

    sage: H.make(2)
    sage: H.gens()
    [1,
     c_2_2: 2-Cocycle in H^*(D8; GF(2)),
     b_1_0: 1-Cocycle in H^*(D8; GF(2)),
     b_1_1: 1-Cocycle in H^*(D8; GF(2))]
    sage: H.rels()
    ['b_1_0*b_1_1']

We look at the generators in more detail. Generator number ``i``
(``i`` an integer) of ``H`` can be obtained by ``H.gen(i)`` or by
``H.i`` (the latter only if ``i`` is a *number*, not just a *variable*
assigned to a number). Note that ``H.0`` always returns the
multiplicative unit in the underlying base field, which is ``GF(2)``
in this case.
::

    sage: H.0
    1
    sage: print H.1
    2-Cocycle in H^*(D8; GF(2)),
    represented by
    [0 0 1]
    rdeg = 1
    ydeg = 0
    sage: print H.2
    1-Cocycle in H^*(D8; GF(2)),
    represented by
    [1 0]
    rdeg = 0
    ydeg = 0
    sage: print H.3
    1-Cocycle in H^*(D8; GF(2)),
    represented by
    [0 1]
    rdeg = 0
    ydeg = 0

``rdeg`` and ``ydeg`` are notions needed in the definition of David
Green's monomial ordering. ``H.1.rdeg()==1`` means that the generator
``H.1`` is Duflot regular. We now compute the restrictions. The
restriction maps can be obtained as follows::

    sage: sorted(H.restriction_maps().items())
    [(1,
      [(2, 1),
       Induced homomorphism of degree 0 from H^*(D8; GF(2)) to H^*(SmallGroup(2,1); GF(2))]),
     (2,
      [(4, 2),
       Induced homomorphism of degree 0 from H^*(D8; GF(2)) to H^*(SmallGroup(4,2); GF(2))]),
     (3,
      [(4, 2),
       Induced homomorphism of degree 0 from H^*(D8; GF(2)) to H^*(SmallGroup(4,2); GF(2))])]

Hence, it is a dictionary (in order to have a reproducible doc test,
we needed to transform it into a sorted list). The keys correspond to
the number of a special subgroup and yield a pair formed by the
SmallGroups identifier of that subgroup and the restriction map. It
can be applied to a generator in multiplicative or functional
notation::

    sage: print H.restriction_maps()[1][1]*H.1
    2-Cocycle in H^*(SmallGroup(2,1); GF(2)),
    represented by
    [1]
    sage: print H.restriction_maps()[1][1](H.1)
    2-Cocycle in H^*(SmallGroup(2,1); GF(2)),
    represented by
    [1]

As required, the restriction of the Duflot regular generator ``H.1``
to `S_1` is not nilpotent. Actually, for `p=2`, the restriction of a
cohomology class to some elementary abelian subgroup is nilpotent if
and only if the restriction is zero.

The list of generators restricted to any subgroup and expressed as a
polynomial in the generators of the subgroup's cohomology ring can be
obtained by
:meth:`~pGroupCohomology.cohomology.COHO.restrict_to_subgroup`::

    sage: H.restrict_to_subgroup(2)
    [c_1_0*c_1_1+c_1_0^2: 2-Cocycle in H^*(SmallGroup(4,2); GF(2)),
     c_1_1: 1-Cocycle in H^*(SmallGroup(4,2); GF(2)),
     0: 1-Cocycle in H^*(SmallGroup(4,2); GF(2))]

Note that the choice of generators and relations depends on how the
group was originally defined, as we have seen above.

Dave Benson's completeness criterion
------------------------------------

Dickson classes
^^^^^^^^^^^^^^^

One key problem is to detect whether all generators and relations are
already known. In the case of an abelian group, it is known that a
computation out to degree 2 will always suffice.  In the case of
generalized quaternion groups, it is known that a computation out to
degree 4 will always suffice. For general prime power groups, we solve
the termination problem according to a method introduced by [Benson]_,
improved by [GreenKing]_, and by another criterion introduced by
[Symonds]_. Note that we use a third criterion in the non prime power
case, as described below.

Let `r` be the `p`-rank of `G`, and let `z` be the rank of the centre
of `G`. Using Dickson invariants, we define `r-z` cohomology classes
`d_{1,j},...,d_{r-z,j}` of `S_j` (`j=2,...,n`) in the polynomial part
of the cohomology ring of `S_j`, restricting to zero on `S_1` (recall
that `S_2,...S_n` are maximal elementary abelian subgroups of `G`, and
`S_1` is a common subgroup, namely the greatest central elementary
abelian subgoup). 

One can show that there exist cohomology classes `D_1,...,D_{r-z}` of
`G` such that `D_i` simultaneously restricts to `d_{i,2},...,d_{i,n}`
or to their `p^{k_i}`-th power for some `k_1,...,k_{r-z}`.  We briefly
refer to these classes as *restricted Dickson classes* or
*restricted Dickson elements* for `G` respectively for `S_j` (or we even
drop the word *restricted*, if it is clear from the context).

Using results of [Benson]_, [GreenKing]_ show that the Duflot regular
generators of `H^*(G)` together with the restricted Dickson classes
for `G` form a filter-regular homogeneous system of parameters (HSOP)
for `H^*(G)`. Filter-regular means that the multiplication map by
parameter number `k` on the quotient of `H^*(G)` modulo the first `k-1`
parameters has finite dimensional kernel. Since they are parameters,
the quotient of `H^*(G)` by all parameters is finite dimensional.

Moreover, by [GreenKing]_, we are allowed to try and factorise the
simultanous lifts of the `d_{i,j}`, possibly after modification by
nilpotent elements. Let `D_i` be the smallest resulting non-constant
factor. And finally, the last parameter may be replaced by any other
parameter, since filter-regularity is automatic for the last parameter.

The motivation for all these refinements is to keep the parameter
degrees low. However, in the case of the dihedral group, we have `n=2`
and the Dickson class of `S_2` is of degree one, thus, is simple
enough.
 
If the Dickson classes for `G` have already been found, there is an
attribute ``Dickson``, which is a list of strings providing a
polynomial representation of the factors of Dickson classes. There are
two widely different ways to find the restricted Dickson elements, see
below. A list of parameters is returned by :meth:`~pGroupCohomology.cohomology.COHO.parameters`.

The dihedral group of order 8 has two non-conjugate maximal elementary
abelian subgroups of order 4. The restriction maps to these subgroups
can be obtained with ``H.restriction_maps()[2]`` and
``H.restriction_maps()[3]``. The restriction map to the greatest
central elementary abelian subgroup can be obtained with
``H.restriction_maps()[1]``.

We verify that the Dickson class of ``H`` really restricts to the
Dickson class of its two special subgroups of order 4, and that the
restriction to the greatest central elementary abelian subgroup is
zero. Note that the restricted Dickson classes of the subgroups are
not available if the optional parameter ``useElimination=True`` is
used (see below). In any case, the restricted Dickson classes of the
subgroups can be constructed using the method
:meth:`~pGroupCohomology.cohomology.COHO.dickson_in_subgroup`::


    sage: H.dickson_in_subgroup((4,2))
    sage: H.subgpDickson[4,2]
    [c_1_1: 1-Cocycle in H^*(SmallGroup(4,2); GF(2))]

During ``H.make(2)``, it was successfully tried to lift it to a
Dickson class in ``H``, using :meth:`~pGroupCohomology.cohomology.COHO.lift_dickson`.
The class is represented by a string::

    sage: H.Dickson
    ['b_1_1+b_1_0']

There is an alternative way of lifting the Dickson classes. While
:meth:`~pGroupCohomology.cohomology.COHO.lift_dickson` works with
linear algebra, :meth:`~pGroupCohomology.cohomology.COHO.find_dickson`
uses elimination. The first method requires that the restriction maps
to the maximal elementary abelian subgroups are known out to the
degree in which the Dickson classes live. The degree grows
exponentially with the difference of the `p`-ranks of the group and its
center and can thus be very high, making this method
unfeasible. Hence, if the degree of the Dickson classes is greater
than 18, the elimination method is used by default. Using the optional
parameter ``useElimination=True``, the elimination method can be
chosen explicitly and can be explicitly avoided by
``useElimination=False``.

In the case of the dihedral group of order 8, both methods work, and
in fact they yield the same result (which is not necessarily the
case)::

    sage: H.Dickson = []
    sage: H.find_dickson()
    sage: H.Dickson
    ['b_1_1+b_1_0']

We verify that the element ``H('b_1_1+b_1_0)`` restricts to zero in
the greatest central elementary abelian subgroup, and restricts to the
Dickson class in the two maximal elementary abelian subgroups::

    sage: print H.restriction_maps()[1][1](H('b_1_1+b_1_0'))
    1-Cocycle in H^*(SmallGroup(2,1); GF(2)),
    represented by
    [0]
    sage: H.subgpDickson[4,2][0] == H.restriction_maps()[2][1](H('b_1_1+b_1_0'))
    True
    sage: H.subgpDickson[4,2][0] == H.restriction_maps()[3][1](H('b_1_1+b_1_0'))
    True

Hence, in our example, we already know that if the cohomology
computation were complete then ``['c_2_2','b_1_0+b_1_1']`` would give
rise to a filter regular system of parameters. Filter regularity can
be checked as follows::

    sage: H.parameters()
    ['c_2_2', 'b_1_1+b_1_0']
    sage: H.raw_filter_degree_type(['c_2_2','b_1_0+b_1_1'])
    ([-1, -1, 1], [[0], [0], [1, 1]], [2, 1])

This answer means that the raw filter degree type is ``[-1, -1, 1]``,
the second list is some data that is useful to compute the Poincare
series, and the last list is the degree list of the given parameters.
The filter degree type are available by
::

    sage: H.filter_degree_type()
    [-1, -2, -2]

The filter degree type together with the degrees of the filter regular 
system of parameters and the rank of the centre of `G` give rise to a 
degree bound of [Benson]_:

**Theorem**

  Let `G` be a finite group. Assume that `R_n` is an approximation of
  `H^\ast(G;\mathbb F_p)` out to degree `n` and assume that there is
  a filter regular HSOP for `R_n` in degrees at least two that gives
  rise to a filter regular HSOP of `H^\ast(G;\mathbb F_p)`. Let
  `f_0,...,f_r` be the filter degree type and let `d_1,...,d_r` be the
  degrees of the filter regular HSOP. If

    `n > \max(0, f_0+0,  f_1+1,...,f_{r-1}+r-1) - r + d_1+ ... +d_r` 

  then `R_n` is isomorphic to the cohomology ring.

Benson conjectures that the filter degree type for any modular
cohomology ring of a finite p-group is of the form
`-1,-2,...,-r,-r`. Note that the strict inequality in the theorem can
be improved to a non-strict inequality if the depth of the cohomology
ring of `G` is of rank at least two; in particular, by a result of
[Duflot]_, this is the case if the centre of a Sylow `p`-subgroup of
`G` is of rank at least two.

How do we apply that theorem? Apparently, the key problem is to find a
filter regular HSOP for the approximation `R_n` of the cohomology
ring. The point is that the Duflot regular generators together with
the Dickson classes form a filter regular HSOP for `H^\ast(G;\mathbb
F_p)`, and that we are able to compute Duflot regular generators and
Dickson classes in `R_n`.

1. Test whether the Duflot regular generators together with the Dickson
   classes form a filter regular HSOP for `R_n`. If it fails, then
   certainly `R_n` is not isomorphic to the cohomology ring. So, we
   need to compute an approximation in higher degrees.
2. If they *do* form a filter regular HSOP for `R_n` then we can compute
   the filter degree type and check whether the inequality in Benson's
   theorem is satisfied.
3. If it *is* satisfied, then `R_n` is isomorphic to the cohomology
   ring, and we can stop. Otherwise, we can not prove that `R_n` is
   isomorphic to the cohomology ring, and thus we must continue in
   higher degrees.

What does it mean, in our example? We have verified that we have 
a filter regular HSOP for the approximation in degree two. But looking
at the degrees of the parameters, Benson's criterion yields the degree 
bound 3. However, it turns out that there is no further generators or 
relations for ``H`` in degree 3. Hence, Benson's criterion is not 
perfect, as it sometimes applies only in higher degrees than necessary.

According to [GreenKing]_, it suffices to have one filter regular hsop
of the ring approximation that gives rise to a hsop of the cohomology
ring, and to prove the existence of a filter regular hsop of smaller
degrees by changing the coefficients to a finite extension field ---
then, one can use Benson's criterion using the degrees of the smaller
hsop (without explicitly constructing it). We elaborate this in the
next section.

Since the filter degree type is independent of the choice of a
filter-regular system of parameters, we can use the Dickson classes
for computing the filter-degree type. Then, we prove the existence of
a filter regular HSOP in smaller degrees. We do not need this HSOP
explicitly --- we only need the degrees (which comes out of the
existence proof) and the filter-degree type (which can be computed
using the Dickson classes).

Peter Symonds' Completeness Criterion
-------------------------------------

**Theorem** [Symonds]_

  Let `R_n` be the approximation of `H^\ast(G;\mathbb F_p)` out to
  degree `n`. Assume that there is a HSOP for `R_n` in degrees
  `d_1,...,d_r` that gives rise to a HSOP of `H^\ast(G;\mathbb
  F_p)`. Let `N` be the maximal degree of generators of `R_n` as a
  module over the parameters. If

    `n \ge \max(N, d_1+ ... +d_r - r +1)` 

  then `R_n` is isomorphic to the cohomology ring.

In order to compute the maximal degree `N` of module generators over
the parametes, it is needed to explicitly know the parameters. Hence,
in contrast to the Benson criterion, it seems impossible to improve
this criterion by an existence result for small parameters over an
extension field.

However, for verifying the Symonds criterion, it is not needed to
compute the filter degree type of the HSOP. Actually, it is not even
needed to have algebraically independent parameters: It suffices to
have elements that generate an ideal of Krull dimension zero.

Since Symonds' criterion is easy to verify, it is used whenever our
construction methods happen to result in sufficiently small
parameters. Otherwise, we try to prove the existence of smaller
parameters over a finite extension field, as in the following section,
and apply a modification of the Benson criterion.


A step-by-step example
----------------------

The purpose of this section is to illustrate what happens behind the scenes
when invoking :meth:`~pGroupCohomology.cohomology.COHO.make`. Of course,
it is not needed to do all these steps manually when computing a cohomology
ring. 

... using a modification of the Benson criterion
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

In our first example, we use group number 32 of order 64.
::

    sage: H = CohomologyRing(64,32,from_scratch=True)

We compute the ring structure out to degree 6. Since we did not use
the optional parameter ``useElimination=True``, some of the Dickson
classes of `G` will be found by the linear algebra method. For the
last one, this would be impossible by the linear algebra method, as
this Dickson class is of degree 7. Note that instead of repeating
``H.next()`` six times, one could also do ``H.make(6)``.
::

    sage: H.next()
    sage: H.next()
    sage: H.next()
    sage: H.next()
    sage: H.next()
    sage: H.next()
    sage: H.Dickson
    ['b_1_1*b_3_6+b_1_1^4+b_2_3^2+b_2_2^2+b_2_1^2',
     'b_1_1^3*b_3_6+b_2_3^2*b_1_1^2+b_2_2*b_1_1*b_3_6+b_2_2*b_2_3^2+b_2_2^2*b_1_1^2+b_2_1*b_2_3^2',
     None]

Although the third Dickson class is in degree 7 and we know the ring
structure only out to degree 6, it is possible to express the third
Dickson class as a polynomial in the generators of our approximation
of the cohomology ring. This fact can not be seen with the linear
algebra method, but it can be seen using elimination, which is done in
:meth:`~pGroupCohomology.cohomology.COHO.find_dickson`.  Note that, by
default, it is tried to factorise the lift of the Dickson classes,
potentially after a modification by some nilpotent element (see
:meth:`~pGroupCohomology.cohomology.COHO.small_factor`). By this little
trick, we actually obtain a parameter in degree one::

    sage: H.find_dickson()
    sage: H.Dickson
    ['b_1_1*b_3_6+b_1_1^4+b_2_3^2+b_2_2^2+b_2_1^2',
     'b_1_1^3*b_3_6+b_2_3^2*b_1_1^2+b_2_2*b_1_1*b_3_6+b_2_2*b_2_3^2+b_2_2^2*b_1_1^2+b_2_1*b_2_3^2',
     'b_1_1']

Since the rank of the centre of the group is one, there is exactly one
Duflot regular generator.  We have already found it, namely in degree
4::

    sage: print H.5
    4-Cocycle in H^*(SmallGroup(64,32); GF(2)),
    represented by
    [0 0 0 0 0 0 0 0 0 0 0 1]
    rdeg = 1
    ydeg = 0
    
Now, as we have the Dickson classes and the Duflot regular generator,
does it make sense to try and apply Benson's criterion? It does not,
since there is one non-regular generator in degree 5::

    sage: H.gens()
    [1,
     b_2_1: 2-Cocycle in H^*(SmallGroup(64,32); GF(2)),
     b_2_2: 2-Cocycle in H^*(SmallGroup(64,32); GF(2)),
     b_2_3: 2-Cocycle in H^*(SmallGroup(64,32); GF(2)),
     b_4_10: 4-Cocycle in H^*(SmallGroup(64,32); GF(2)),
     c_4_11: 4-Cocycle in H^*(SmallGroup(64,32); GF(2)),
     a_1_0: 1-Cocycle in H^*(SmallGroup(64,32); GF(2)),
     b_1_1: 1-Cocycle in H^*(SmallGroup(64,32); GF(2)),
     b_3_4: 3-Cocycle in H^*(SmallGroup(64,32); GF(2)),
     b_3_5: 3-Cocycle in H^*(SmallGroup(64,32); GF(2)),
     b_3_6: 3-Cocycle in H^*(SmallGroup(64,32); GF(2)),
     b_5_17: 5-Cocycle in H^*(SmallGroup(64,32); GF(2))]

We thus expect that there will be a relation at least in degree
2*5=10. This heuristical lower bound for the necessary degree of
computations is available by a method::

    sage: H.expect_last_relation()
    10

So, we continue out to degree 10. Indeed, we find a relation in that
degree, but there are no further generators. Now it makes sense to
test whether the Duflot regular generator and the lifts of the Dickson
classes form a filter regular homogeneous system of parameters::

    sage: H.next()
    sage: H.next()
    sage: H.next()
    sage: H.next()
    sage: H.lastRel
    10
    sage: H.raw_filter_degree_type(['c_4_11']+H.Dickson)
    ([-1, -1, 3, 10, 11], [[0], [0], [0, 1, 0, 1], [0, 1, 1, 2, 2, 2, 2, 2, 1, 1, 1], [1, 1, 3, 4, 5, 6, 5, 6, 3, 3, 1, 1]], [4, 4, 6, 1])

The Filter Degree Type is in accordance with Benson's Strong
Regularity Conjecture::
    
    sage: H.filter_degree_type()
    [-1, -2, -3, -4, -4]

The original form of Benson's criterion would only apply in a very high degree:

- The `p`-rank of the group is 4. Benson originally suggested to construct
  parameters by lifting Dickson elements in the maximal elementary abelian
  subgroups (*not* restricted to a complement of the greatest central
  elementary abelian subgroup). These unrestricted Dickson classes live
  in degrees 8, 12, 14 and 15, or powers thereof.
- The centre is of rank less than two, so, we can not be sure that the
  depth of the cohomology ring is at least two.
- By consequence, Benson's criterion applies, in the best case, in
  degree `7+11+13+14+1=46`.
 
However, by our improved choice of filter-regular parameters, we obtain
the degree bound `3+3+5+1+1=13`. But we can still do better.

**Theorem** [GreenKing]_

    Let `p_1,...,p_r` be a filter-regular HSOP for the approximation
    `R_n` of `H^\ast(G;\mathbb F_p)` that gives rise to a filter
    regular HSOP for the cohomology ring. Assume that for some `m\le r`,
    the quotient of `R_n` modulo `p_1,...,p_m` and modulo all monomials
    of some degree `d` is finite.

    Then, there is a filter-regular HSOP for the degree `n`
    approximation of `H^\ast(G;K)` for some finite field extension
    `K` of `\mathbb F_p`, obtained from `p_1,...,p_m` and `r-m`
    elements of degree `d`.

    Moreover, the degrees of this HSOP of `H^\ast(G;K)` can be used
    as in Benson's criterion to prove completeness of the ring
    approximation of `H^\ast(G;\mathbb F_p)`.

This result is used in
:meth:`~pGroupCohomology.cohomology.COHO.potential_degree_bound`::

    sage: H.potential_degree_bound(['c_4_11']+H.Dickson, [4,4,6,1])
    7

That means: If it is true that ``['c_4_11']+H.Dickson`` is a
filter-regular HSOP of ``H`` of an appropriate filter degree type and
if we happen to know the ring structure is known at least out to
degree 7, then the computation is complete. Since we already verified
filter regularity and since we know the ring structure not only to
degree 7 but to degree 10, we are done.

We remark that we only need to know the mere *existence* of a
filter-regular HSOP in low degrees *over a finite field extension*.
It is not needed, and in general impossible, to construct the small
parameters without extending the field. However, here it is possible,
and we do so, just for illustration. 

The attribute ``WhatFRS`` tells us that we can replace the last three 
parameters (which are of degree 4, 6, and 7) by elements in degree 2::

    sage: H.WhatFRS
    (3, 2)

We test whether the three generators in degree two can be
used as parameters::

    sage: H.raw_filter_degree_type(['c_4_11','b_2_1','b_2_2','b_2_3'])

This is not the case, so, we try some other elements::

    sage: H.raw_filter_degree_type(['c_4_11', 'b_2_1+b_2_2', 'b_2_1+b_2_3', 'b_2_2+b_2_3'])
    sage: H.raw_filter_degree_type(['c_4_11', 'b_2_1+b_2_2', 'b_1_1^2+b_2_3', 'b_2_2+b_2_3'])
    ([-1, -1, 1, 4, 6], [[0], [0], [0, 1], [0, 1, 1, 1, 1], [1, 2, 1, 3, 3, 1, 1]], [4, 2, 2, 2])

So, indeed, it is possible (but not needed) to explicitly provide
three filter regular parameters in degree two, in addition to the
Duflot regular generator.

... using the Symonds criterion
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

**Theorem** [Symonds]_

    Assume that the cohomology ring is known out to degree `n`, and
    assume that the ring approximation has a finite subset `P` in
    degrees `d_1,...,d_r` that gives to an ideal of Krull dimension
    zero in the cohomology ring. If `n > \sum_{i=1}^r (d_i-1)` and if
    the ring approximation is generated in degree at most `n` as a
    module over the algebra generated by `P`, then the ring
    approximation is complete.

Note that it is not assumed that `P` is filter regular. This gives
more freedom. The strategy in :meth:`~pGroupCohomology.cohomology.COHO.parameters`
is to start with a filter regular hsop, which can be constructed
using Duflot generators and Dickson elements as explained above,
see :meth:`~pGroupCohomology.cohomology.COHO.filter_regular_parameters`.
Then, it is tested by exhaustive search whether any of these
filter regular parameters can be replaced by a parameter of smaller 
degrees. This will, in general, result in a hsop that is not 
filter regular, but has smaller degrees.

In our example, we indeed obtain parameters in smaller degrees::

    sage: H.filter_regular_parameters()
    ['c_4_11',
     'b_1_1*b_3_6+b_1_1^4+b_2_3^2+b_2_2^2+b_2_1^2',
     'b_1_1^3*b_3_6+b_2_3^2*b_1_1^2+b_2_2*b_1_1*b_3_6+b_2_2*b_2_3^2+b_2_2^2*b_1_1^2+b_2_1*b_2_3^2',
     'b_1_1']
    sage: H.parameters()
    ['c_4_11', 'b_2_3+b_2_2+b_2_1', 'b_2_2+b_2_1', 'b_1_1']

In this particular case, the small hsop is filter regular, but
we can not rely on it. Anyway, we have `10>4+2+2+1-4`, and it
turns out that the Symonds test applies, since in fact the
ring approximation is generated in degree at most 5 over the
parameters::

    sage: H.SymondsTest(H.parameters(),[4,2,2,1])
    True
    sage: H._max_module_deg
    5

There are cases in which the test for filter regularity can be
quite expensive. On the other hand, it can be very expensive to
compute a minimal resolution in needlessly high degree. Hence,
*if* Symonds' criterion has a chance to apply then we try this
first. The modified Benson test is only used if it would yield 
a better degree bound, due to the existence of smaller parameters
over an extension field. Here, the Symonds test is used. This
can be seen in the protocol (after resetting an internal parameter)::

    sage: H.suffDeg=-1
    sage: H.option('prot')
    sage: H.test_for_completion()
        Try to find a set of generators over which the ring is finite.
          > Considering 10 elements
          > Considering 9 elements
          > Considering 9 elements
          > Considering 8 elements
          > Considering 7 elements
          > Considering 6 elements
          > Considering 5 elements
          > Considering 5 elements
          > Considering 5 elements
        Trying Symonds' criterion for H^*(SmallGroup(64,32); GF(2))
        Successful application of the Symonds criterion
    True
    sage: H.option('noprot')


3. Background for the Cohomology Computation of Non Prime Power Groups
======================================================================

In this section, we explain the computation of the modular cohomology
of a finite group that is not of prime power order. We do it by means
of an example that we compute step by step. Of course, invoking
:meth:`~pGroupCohomology.modular_cohomology.MODCOHO.make` will do all
these computations behind the scenes.

Our main example is the mod-2 cohomology of the Symmetric Group on
8 elements.
::

    sage: G = gap('SymmetricGroup(8)')
    sage: tmp = tmp_filename()
    sage: CohomologyRing.set_user_db(tmp)
    sage: HS8 = CohomologyRing(G, prime=2, GroupName='SymmetricGroup(8)', from_scratch=True)


The Stable Element Method
-------------------------

For the theoretical background of the Stable Element Method, we refer to [AdemMilgram]_.
It is known that the mod-`p` cohomology ring of a finite group `G` is
a sub-ring of the mod-`p` cohomology of any subgroup `P` whose index in `G`
is co-prime to `p`. This sub-ring is formed by those cocycles of `P`
that are stable under conjugation by elements of `G`. That means, a
cocycle `c\in H^\ast(P; \mathbb F_p)` is stable if and only if for all
`g\in G` holds `r_{P, P\cap P^g}(c) = r_{P^g, P\cap P^g}(\phi_g(c))`,
where `\phi_g` denotes the homomorphism induced by conjugation with `g^{-1}`,
and `r_{P, P\cap P^g}, r_{P^g, P\cap P^g}` are restriction maps.

It suffices to test stability for representatives of all double cosets of `P`
in `G`. While this is, of course, much better than testing all elements of `G`,
the number of double cosets can still be rather large. It is thus often
better to compute the cohomology of `G` in two steps. By default, we proceed as
follows.

 - Let `S` be a Sylow `p`-subgroup of `G`. Compute its mod-`p` cohomology,
   ``HS``.
 - Let `N = N_G(Z(S))`. Compute its mod-`p` cohomology ``HN`` as a sub-ring
   of ``HS``.
 - If `N\not=G`: Compute the mod-`p` cohomology of `G` as a sub-ring of ``HN``.
    
In our example, the subgroups look as follows::

    sage: HS8.subgroup_cohomology() # HS8 is a sub-ring of HS8.subgroup_cohomology()
    H^*(SmallGroup(384,5602); GF(2))
    sage: HS8.subgroup_cohomology().subgroup_cohomology()
    H^*(Syl2(S8); GF(2))
    sage: HS8.subgroup_cohomology().subgroup_cohomology() is HS8.sylow_cohomology()
    True

The stability conditions are expressed by pairs of induced
homomorphisms starting at ``HS8.subgroup_cohomology()``::

    sage: HS8._PtoPcapCPdirect
    [Induced homomorphism of degree 0 from H^*(SmallGroup(384,5602); GF(2)) to H^*(D8xV4; GF(2)), Induced homomorphism of degree 0 from H^*(SmallGroup(384,5602); GF(2)) to H^*(SmallGroup(32,27); GF(2)), Induced homomorphism of degree 0 from H^*(SmallGroup(384,5602); GF(2)) to H^*(D8; GF(2))]
    sage: HS8._PtoPcapCPtwist
    [Induced homomorphism of degree 0 from H^*(SmallGroup(384,5602); GF(2)) to H^*(D8xV4; GF(2)), Induced homomorphism of degree 0 from H^*(SmallGroup(384,5602); GF(2)) to H^*(SmallGroup(32,27); GF(2)), Induced homomorphism of degree 0 from H^*(SmallGroup(384,5602); GF(2)) to H^*(D8; GF(2))]

The elements of ``HS8`` are given by those elements of ``HS8.subgroup_cohomology()``
for which the images coincide for each pair. These so-called *stable elements*
can be computed in a specific degree by
:meth:`~pGroupCohomology.modular_cohomology.MODCOHO.stable_space`,
which also returns a list of standard monomials for ``HS8.subgroup_cohomology()``
and a sublist of those monomials that occur as leading monomials in
stable cocycles::

    sage: HS8.stable_space(2)
    ([Gen2: 2-Cocycle in H^*(SymmetricGroup(8); GF(2)), Gen1: 2-Cocycle in H^*(SymmetricGroup(8); GF(2))], ['b_1_1^2', 'b_1_0*b_1_1', 'b_1_0^2', 'b_2_4', 'b_2_3'], ['b_1_1^2', 'b_1_0^2'])


Construction of filter regular parameters
-----------------------------------------

In the case of prime power groups, we form a filter regular hsop out
of a Duflot regular sequence (choosing the parameters such that a
Duflot regular sequence is formed by generators), together with
simultaneous lifts of Dickson elements in the maximal elementary
abelian subgroups in a complement of the greatest central elementary
abelian subgroup. For general finite groups, we need to use slightly
different methods.

Duflot regular sequence
^^^^^^^^^^^^^^^^^^^^^^^

In our main example ``HS8``, the rank of the centre of a Sylow
subgroup is one. Hence, any Duflot element yields a maximal Duflot
regular sequence. This is found in degree four.
::

    sage: HS8.make(4)
    sage: HS8.duflot_regular_sequence()
    ['c_4_0']

There are some difficulties for non prime power groups, though. For
starters, the number of generators with non-trivial restriction to the
greatest central elementary abelian subgroup of a Sylow `p`-subgroup
can exceed the rank of the greatest central `p`-elementary abelian
subgroup. In the following example, all 5 generators are Duflot
elements, but only three of them form a regular sequence::

    sage: H96 = CohomologyRing(96,3,prime=2)
    sage: H96.make()
    sage: [t for t in H96.gens()[1:] if t.rdeg()]
    [c_2_0: 2-Cocycle in H^*(SmallGroup(96,3); GF(2)), c_4_3: 4-Cocycle in H^*(SmallGroup(96,3); GF(2)), c_6_4: 6-Cocycle in H^*(SmallGroup(96,3); GF(2)), c_6_6: 6-Cocycle in H^*(SmallGroup(96,3); GF(2)), c_6_7: 6-Cocycle in H^*(SmallGroup(96,3); GF(2))]
    sage: H96.duflot_regular_sequence()
    ['c_2_0', 'c_4_3', 'c_6_7']

Moreover, sometimes the program will only find to a Duflot regular
sequence that is not entirely formed by generators, as in the
following example::

    sage: H48 = CohomologyRing(48,50, prime=2)
    sage: H48.make()
    sage: H48.duflot_regular_sequence()
    ['c_2_3', 'c_2_2', 'c_3_1', 'c_3_7+(c_3_0)']


Simultaneous lifts of Dickson elements
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

By [Benson]_, it is always possible to find elements in the cohomology
of `G` that restrict to (powers of) Dickson elements in all maximal
elementary abelian subgroups, and these elements form a filter regular
hsop.

The degrees of these elements grow exponentially in the `p`-rank of
`G`. This is not good, since all known completeness criteria depend on
the degrees of the parameters. However, Benson's construction
certainly is a reasonable starting point, that we refine by various
methods.

Dickson elements in a complement of the centre of a Sylow subgroup
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Above, we stated that a filter regular hsop results if one extends a
maximal Duflot regular sequence by simultaneous lifts of Dickson
elements of complements of the greatest central elementary abelian
subgroup of a Sylow subgroup in maximal elementary abelian
subgroups. This holds in general.

However, the *existence* of the simultaneous lifts is only granted
for prime power groups, by [GreenKing]_. For groups that are not of
prime power order, they do not exist, in general.

But often enough, the simultaneous lifts *do* exist. Therefore, we
proceed as follows.

    * Since we use the stable element method, the cohomology ring
      of a Sylow subgroup `S` is already known when we start the 
      cohomology computation of a finite group `G`.
    * Simultaneous lifts of Dickson elements in maximal elementary 
      abelian subgroups in a complement of the greatest central 
      elementary abelian subgroup of `S` exist in the cohomology
      of `S`.
    * If these simultaneous lifts happen to be stable, then they
      yield elements of the cohomology of `G`. Together with a
      maximal Duflot regular sequence of the cohomology of `G`,
      they yield a filter regular hsop.

In the following example, the `p`-rank of the group is four, hence,
the Dickson elements are in degrees 8, 12, 14 and 15. But the
centre of a Sylow subgroup is of rank two, and therefore, the
Dickson elements in a complement of the centre of a Sylow subgroup
are of degree 2 and 3. In fact, the degree 3 element can even
be replaced by a parameter in degree 1, for the Sylow subgroup::

    sage: H576 = CohomologyRing(576,5000, prime=2)
    sage: PS = H576.sylow_cohomology().filter_regular_parameters(); PS
    ['c_4_50', 'c_1_4', 'b_1_0^2 + b_1_0*b_1_1 + b_1_1^2 + b_1_0*b_1_2 + b_1_2^2 + b_1_1*b_1_3 + b_1_2*b_1_3 + b_1_3^2', 'b_1_1 + b_1_2']

It turns out that the above filter regular parameters for the
Sylow subgroup are stable::

    sage: H576.make()
    sage: H576.parameters_from_sylow_subgroup()
    ['c_1_2', 'c_4_12', 'b_1_1^2+b_1_0*b_1_1+b_1_0^2', 'b_1_1']

If the last parameter for the cohomology of the Sylow subgroup were not
stable, we could use :meth:`~pGroupCohomology.cohomology.COHO.find_small_last_parameter`
to find another parameter, which is done internally in
:meth:`~pGroupCohomology.cohomology.COHO.filter_regular_parameters`.

For ``H96`` (an example that we started above), the rank of a maximal
2-elementary abelian subgroup is three. So, the Dickson elements are
of degree 4, 6 and 7. This would only give rise to a degree bound of
3+5+6=14. Nevertheless, the Duflot regular sequence provides a better
filter-regular HSOP, and indeed the program automatically replaces the
Dickson elements by the Duflot generators::

    sage: H96.filter_regular_parameters()
    ['c_2_0', 'c_4_3', 'c_6_7']

Just for demonstration, we reconstruct the original Dickson elements,
without any simplification::

    sage: H96.setprop('useFactorization',False)
    sage: H96.find_dickson()
    sage: H96.Dickson
    ['c_4_3^2+c_2_0*c_6_7+c_2_0^2*c_4_3+c_2_0^4', 'c_6_4*c_6_7+c_6_4^2+c_4_3^3+c_2_0^2*c_4_3^2', 'c_2_0*c_6_4*c_6_7+c_2_0*c_6_4^2+c_2_0*c_4_3^3+c_2_0^4*c_6_7+c_2_0^5*c_4_3']


Finding smaller parameters
^^^^^^^^^^^^^^^^^^^^^^^^^^

In [GreenKing]_, we show that the last element of a filter regular
hsop can always be replaced by any other parameter --- the result will
still be a filter regular hsop. The method
:meth:`~pGroupCohomology.cohomology.COHO.find_small_last_parameter`
tries to find a parameter of minimal degree.

However, if this replacement is not done in the last position, the
resulting hsop is not necessarily filter regular. But both for the
Symonds and the Hilbert-Poincar\'e criterion explained below, a
general hsop suffices. In the following example, a hsop can be
constructed in much smaller degrees than a filter regular hsop.
Since some randomised algorithms of Gap are used in the background,
we only give the degrees of the parameters::

    sage: H = CohomologyRing(384,18136,prime=2, from_scratch=True)
    sage: H.make()
    sage: H.set_ring()
    sage: [singular('deg(%s)'%p) for p in H.filter_regular_parameters()]
    [8, 12, 14, 3]
    sage: [singular('deg(%s)'%p) for p in H.parameters()]
    [8, 12, 2, 3]

These parameters are not guaranteed to be filter regular, and are thus of no use
for the Benson test. By coincidence, in this example, they are filter regular::

    sage: H.raw_filter_degree_type(H.parameters())
    ([-1, -1, -1, 18, 21], [[0], [0], [0], [0, 1, 1, 2, 2, 3, 3, 4, 4, 4, 4, 4, 4, 3, 3, 2, 2, 1, 1], [1, 1, 1, 3, 6, 5, 4, 6, 10, 9, 6, 6, 9, 9, 5, 3, 4, 5, 2, 0, 0, 1]], [8, 12, 2, 3])

We return to our main example ``HS8``. In this example, the parameters
for the cohomology of the Sylow subgroup are not stable. The last Dickson
element for ``HS8`` is of degree 15. But we don't need to compute that far,
since the Dickson elements can be expressed by generators in smaller
degrees.
::

    sage: HS8.filter_regular_parameters()
    sage: HS8.make(7)
    sage: HS8.parameters_from_sylow_subgroup()
    ['c_4_0', None, None, None]
    sage: HS8.option('prot')
    sage: HS8.generator_degbound()
    New generators for H^*(SymmetricGroup(8); GF(2)) will be found at most in degree 8
    sage: HS8.option('noprot')
    sage: HS8.find_dickson()
    sage: P = HS8.Dickson
    sage: HS8.set_ring()
    sage: [singular('deg(%s)'%p) for p in P]
    [8, 12, 14, 15]

As the ring approximation is carried to higher degrees, we can eventually
replace the last parameter by an element of degree 6::

    sage: HS8.make(14)
    sage: HS8.find_small_last_parameter(P)
    True
    sage: [singular('deg(%s)'%p) for p in P]
    [8, 12, 14, 6]

The above construction is used in :meth:`~pGroupCohomology.cohomology.COHO.filter_regular_parameters`,
and so we obtain
::

    sage: P == HS8.filter_regular_parameters()
    True

The modified Benson criterion will only apply in rather high degree::

    sage: HS8.potential_degree_bound(P,[8,12,14,6])
    36

By enumerative search, we find that we can replace several parameters
by elements of smaller degree. However, there is no guarantee that
this hsop is filter regular. So, it is of no use for the Benson
criterion, but may be handy in the Symonds criterion.
::

    sage: [singular('deg(%s)'%p) for p in HS8.parameters()]
    [4, 6, 7, 6]


A completeness criterion for non prime power groups
---------------------------------------------------

Above, we explained how Benson's and Symonds' completeness criteria
work. For non prime power groups, a criterion is available that
combines the advantages of both. It is formed by two separate
completeness criteria for generators and relations. See [King]_
for more details.

Bounding the generator degree
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

It is well known that `H^\ast(P; \mathbb F_p)` is a finitely generated
`H^\ast(G; \mathbb F_p)`-module via the restriction from `G` to `P`.
Using the fact that transfer is a module homomorphism, one can easily
show that the maximal degree of a generator of `H^\ast(G; \mathbb F_p)`
is bounded by the generator degree of `H^\ast(P; \mathbb F_p)` as a
`H^\ast(G; \mathbb F_p)`-module. This can also be computed in an
approximation of `H^\ast(G; \mathbb F_p)`, and is implemented in
:meth:`~pGroupCohomology.modular_cohomology.MODCOHO.generator_degbound`.

This method was automatically invoked when we computed the ring structure
out to degree 14. Two attributes provide the result::

    sage: HS8.degbound_for_gens
    8
    sage: HS8.all_generators_found
    True

Hence, the program knows that all stable elements above degree 8 can be 
expressed as polynomials of the generators of the ring approximation. It 
is then not needed to compute the stable subspace by means of 
:meth:`~pGroupCohomology.modular_cohomology.MODCOHO.stable_space`, which 
saves a considerable amount of resources.

The Hilbert-Poincare criterion
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

There remains to determine whether all relations have been found.
The following result arose in a discussion of Simon King with
Peter Symonds. Recall that we compute `H^\ast(G; \mathbb F_p)`
as a subring of `H^\ast(P; \mathbb F_p)` for a subgroup `P<G`.

**Theorem**

    Assume that there is a finite field extension `K` over `\mathbb F_p`
    so that `H^\ast(G; K)` has a hsop `P` of degrees
    `d_1,...,d_r`. Let `N = \sum_{i=1}^r d_i - \mathrm{depth}H^\ast(P; \mathbb F_p)`.
    Let `R` be the degree-`n` approximation of `H^\ast(G; \mathbb F_p)`,
    with `n\ge N`, and assume that it contains all generators of
    `H^\ast(G; \mathbb F_p)`. 

    Then, `R` is isomorphic to `H^\ast(G; \mathbb F_p)`
    if and only if the Poincar\'e series of `R` multiplied with
    `\prod_{i=1}^r(1-t^{d_i})` is a polynomial of degree `\le N`.

We explained above how parameters for `H^\ast(G; \mathbb F_p)` can be
constructed in a ring approximation. We found parameters in degrees 4,
6, 7 and 6. Let us test whether we can show the existence of smaller
parameters over a finite extension field::

    sage: HS8.option('prot')
    sage: HS8.HilbertPoincareTest()
    Try to find small parameters
    We expect that the Hilbert-Poincare criterion will not apply before degree 20
    sage: HS8.WhatFRS
    sage: HS8.option('noprot')
    sage: HS8.subgroup_cohomology().depth()
    3

The attribute ``WhatHSOP`` tells us whether it is possible to replace the
parameters provided above by even smaller parameters, after extending the
field. Apparently this is not the case, so, we will work without existence
proof. Since the depth of the cohomology of the subgroup is three,
there is a chance for the criterion to work in degree `20 = 4+6+7+6-3`.
We now complete the computation and verify the criterion manually::

    sage: HS8.make(20)
    sage: p = HS8.poincare_series()
    sage: t = p.parent().gen(0)
    sage: p*(1-t^4)*(1-t^6)*(1-t^7)*(1-t^6)
    t^19 + t^18 + 2*t^17 + 4*t^16 + 5*t^15 + 8*t^14 + 9*t^13 + 11*t^12 + 14*t^11 + 14*t^10 + 14*t^9 + 14*t^8 + 12*t^7 + 10*t^6 + 8*t^5 + 5*t^4 + 4*t^3 + 2*t^2 + t + 1

Indeed, the result is a polynomial of degree at most 20. Therefore, the
computation is complete.

Comparison of the criteria
^^^^^^^^^^^^^^^^^^^^^^^^^^

As we have seen, the Hilbert-Poincare criterion applies in degree 20.
For applying the Symonds criterion, we can use the same parameters,
which are of degrees 4, 6, 7 and 6. This means that the Symonds
criterion applies in degree `4+6+7+6-3=20`, like the Hilbert-Poincare
criterion.

The best filter regular hsop that our automatic methods can provide
has degrees 8, 12, 14 and 6::

    sage: [singular('deg(%s)'%p) for p in HS8.filter_regular_parameters()]
    [8, 12, 14, 6]

It turns out that even in an extension field it is not possible to
replace the last few parameters in smaller degrees. Hence, the
modified Benson test only applies in degree `8+12+14+6-4=36`.

The Hilbert-Poincar\'e criterion combines the advantage of the
modified Benson criterion (using the existence result for small
parameters over an extension field) with the advantage of the Symonds
criterion (the conditions in the criterion are easy to verify, and it
is possible to use algebraically dependent parameters).

In this example, the Symonds criterion and the Hilbert-Poincare
criterion are of equal strength and are much better than the modified
Benson criterion. However, the situation is different in different
examples. See :meth:`pGroupCohomology.cohomology.COHO.test_for_completion`,
:meth:`pGroupCohomology.modular_cohomology.MODCOHO.test_for_completion`,
:meth:`~pGroupCohomology.cohomology.COHO.BensonTest`,
:meth:`~pGroupCohomology.cohomology.COHO.SymondsTest`, or
:meth:`~pGroupCohomology.modular_cohomology.MODCOHO.HilbertPoincareTest`.

References
==========

.. [AdemMilgram] Alejandro Adem, R. James Milgram: *Cohomology of finite groups.* Second edition. Grundlehren der Mathematischen Wissenschaften [Fundamental Principles of Mathematical Sciences], **309.** Springer-Verlag, Berlin, 2004.
.. [Benson] Dave Benson: Dickson invariants, regularity and computation in group cohomology. *Illinois J. Math.* **48** (2004), 171--197.
.. [BensonCarlson] Dave Benson, Jon F. Carlson: Projective resolutions and Poincar\'e duality complexes. *Trans. Amer. Math. Soc.* **342** (1994), 447--488.
.. [CartanEilenberg] Henri Cartan and Samuel Eilenberg: *Homological algebra*. Princeton University Press, Princeton, N. J., 1956.
.. [Duflot] Jeanne Duflot: Depth and equivariant cohomology, *Comment. Math. Helv.* **56** (1981), 627--637.
.. [EllisKing] Graham Ellis and Simon King: `Persistent homology of groups. <http://arxiv.org/abs/1006.2237>`_, to appear in *J. Group Theory*.
.. [Green] David Green: Gr\"obner bases and the computation of group cohomology.  *Lecture Notes in Mathematics,* **1828**. Springer-Verlag, Berlin, 2003.
.. [GreenKing] David Green, Simon King: `The computation of the cohomology rings of all groups of order 128. <http://arxiv.org/abs/1001.2577>`_, to appear in J. Algebra.
.. [King] Simon King: `Completeness criteria for modular cohomology rings of non prime power groups. <http://arxiv.org/abs/1004.0736>`_, *Preprint,* 2010.
.. [KingGreenEllis] Simon King, David Green, Graham Ellis: `The Mod-2 Cohomology Ring of the Third Conway Group is Cohen-Macaulay. <http://arxiv.org/abs/1005.2472>`_, *Preprint,* 2010.
.. [Kraines] David Kraines: Massey Higher Products. *Trans. AMS* **124** (1966), 431--449.
.. [Kuhn] Nicholas J. Kuhn: Primitives and central detection numbers in group cohomology, *Adv. Math.* **216** (2007), 387--442.
.. [Pham] Pham Anh Minh: Modular invariant theory and cohomology algebras of extra-special `p`-groups. *Pacific J. Math.* **124** (1986), 345--363.
.. [Ravenel] Douglas Ravenel, *Complex cobordism and stable homotopy groups of spheres,* Pure and Applied Mathematics, **121**. Academic Press, Inc., Orlando, FL, (1986).
.. [Symonds] Peter Symonds: `On the Castelnuovo-Mumford Regularity of the Cohomology Ring of a Group <http://www.maths.manchester.ac.uk/~pas/preprints/>`_.  *J. American Math. Soc.,* **23** (2010). 
.. [Wilkerson] Clarence Wilkerson: *A primer on the Dickson invariants,* Proceedings of the Northwestern Homotopy Theory Conference (Evanston, Ill., 1982), *Contemp. Math.,* **19** (1983), 421--434.


"""

from pGroupCohomology import barcode, factory
from pGroupCohomology.factory import CohomologyRing
from pGroupCohomology.resolution import gap, singular
