#*****************************************************************************
#
#    Copyright (C) 2009/10 Simon A. King <simon.king@nuigalway.ie>
#
#  Distributed under the terms of the GNU General Public License (GPL),
#  version 2 or later (at your choice)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext
from sage.all import SAGE_ROOT
from sage.interfaces.singular import singular
from sage.all import srange
from sage.rings.integer import Integer


import shutil
import os
import subprocess

# Copy our singular libraries to the right place

try:
    SingularLib=os.environ['SINGULARPATH']
except KeyError:
    SingularLib=singular.eval('system("SingularLib")').split(':')[0]
shutil.copy('pGroupCohomology/dickson.lib',SingularLib)
shutil.copy('pGroupCohomology/filterregular.lib',SingularLib)

PrimePowers = [n for n in srange(2,256) if Integer(n).is_prime_power()]
os.chdir('lib/')
for n in PrimePowers:
    subprocess.call(['../bin/maketab','%d'%n])
os.chdir('../')

mtx_tables = ['lib/'+s for s in os.listdir('lib/') if s.endswith('zzz')]

setup(
  name = "pGroupCohomology",
  version = "2.1.1",
  author = "Simon A. King, David J. Green",
  author_email = "simon.king@nuigalway.ie, david.green@uni-jena.ie",
  maintainer = "Simon A. King", 
  maintainer_email = "simon.king@nuigalway.ie",
  url = "http://sage.math.washington.edu/home/SimonKing/Cohomology/",
  description = "Modular Cohomology Rings of Finite Groups",
  packages=["pGroupCohomology"],
  py_modules = ["pGroupCohomology.barcode","pGroupCohomology.factory"],
  data_files=[("pGroupCohomology",
              ["pGroupCohomology/GroupNames.sobj",
               "pGroupCohomology/GapMaxels",
               "pGroupCohomology/GapMB",
               "pGroupCohomology/GapSgs"]),
              ("pGroupCohomology", ['db/'+X for X in os.listdir('db/')]+["db/LICENCE"]),
              ("pGroupCohomology/mtx_tables", mtx_tables)],
  ext_modules=[ 
    Extension("pGroupCohomology.resolution",
              sources = ["pGroupCohomology/resolution.pyx",
                         "pGroupCohomology/c_sources/aufloesung.c",
                         "pGroupCohomology/c_sources/aufnahme.c",
                         "pGroupCohomology/c_sources/djgerr.c",
                         "pGroupCohomology/c_sources/fileplus.c",
                         "pGroupCohomology/c_sources/nBuchberger.c",
                         "pGroupCohomology/c_sources/pgroup.c",
                         "pGroupCohomology/c_sources/pincl.c",
                         "pGroupCohomology/c_sources/slice.c",
                         "pGroupCohomology/c_sources/urbild.c",
                         "pGroupCohomology/c_sources/uvr.c"],
              libraries = ["mtx","csage"],
              #extra_compile_args=["-g"],
              #extra_link_args=["-g"],
              include_dirs = [SAGE_ROOT+"/local/include/csage/", SAGE_ROOT+"/devel/sage/", SAGE_ROOT+"/devel/sage/c_lib", SAGE_ROOT+"/devel/sage/sage/ext", "mtx2.2.4/src/", "pGroupCohomology/c_sources", "pGroupCohomology"]
              ),
    Extension("pGroupCohomology.cohomology",
              sources = ["pGroupCohomology/cohomology.pyx",
                         "pGroupCohomology/c_sources/aufloesung.c",
                         "pGroupCohomology/c_sources/aufnahme.c",
                         "pGroupCohomology/c_sources/djgerr.c",
                         "pGroupCohomology/c_sources/fileplus.c",
                         "pGroupCohomology/c_sources/nBuchberger.c",
                         "pGroupCohomology/c_sources/pgroup.c",
                         "pGroupCohomology/c_sources/pincl.c",
                         "pGroupCohomology/c_sources/slice.c",
                         "pGroupCohomology/c_sources/urbild.c",
                         "pGroupCohomology/c_sources/uvr.c"],
              libraries = ["mtx"],
              include_dirs = [SAGE_ROOT+"/local/include/csage/", SAGE_ROOT+"/devel/sage/", "mtx2.2.4/src/", "pGroupCohomology/c_sources", "pGroupCohomology"]
              ),
    Extension("pGroupCohomology.modular_cohomology",
              sources = ["pGroupCohomology/modular_cohomology.pyx",
                         "pGroupCohomology/c_sources/aufloesung.c",
                         "pGroupCohomology/c_sources/aufnahme.c",
                         "pGroupCohomology/c_sources/djgerr.c",
                         "pGroupCohomology/c_sources/fileplus.c",
                         "pGroupCohomology/c_sources/nBuchberger.c",
                         "pGroupCohomology/c_sources/pgroup.c",
                         "pGroupCohomology/c_sources/pincl.c",
                         "pGroupCohomology/c_sources/slice.c",
                         "pGroupCohomology/c_sources/urbild.c",
                         "pGroupCohomology/c_sources/uvr.c"],
              libraries = ["mtx"],
              include_dirs = [SAGE_ROOT+"/local/include/csage/", SAGE_ROOT+"/devel/sage/", "mtx2.2.4/src/", "pGroupCohomology/c_sources", "pGroupCohomology"]
              ),
    Extension("pGroupCohomology.cochain",
              sources = ["pGroupCohomology/cochain.pyx",
                         "pGroupCohomology/c_sources/aufloesung.c",
                         "pGroupCohomology/c_sources/aufnahme.c",
                         "pGroupCohomology/c_sources/djgerr.c",
                         "pGroupCohomology/c_sources/fileplus.c",
                         "pGroupCohomology/c_sources/nBuchberger.c",
                         "pGroupCohomology/c_sources/pgroup.c",
                         "pGroupCohomology/c_sources/pincl.c",
                         "pGroupCohomology/c_sources/slice.c",
                         "pGroupCohomology/c_sources/urbild.c",
                         "pGroupCohomology/c_sources/uvr.c"],
              libraries = ["mtx","csage"],
              include_dirs = [SAGE_ROOT+"/local/include/csage/", SAGE_ROOT+"/devel/sage/", SAGE_ROOT+"/devel/sage/c_lib", SAGE_ROOT+"/devel/sage/sage/ext", "mtx2.2.4/src/", "pGroupCohomology/c_sources", "pGroupCohomology"]
              ),
    Extension("pGroupCohomology.mtx",
              sources = ["pGroupCohomology/mtx.pyx"],
              libraries = ["mtx","csage"],
              #extra_compile_args=["-g"],
              #extra_link_args=["-g"],
              include_dirs = [SAGE_ROOT+"/local/include/csage/", SAGE_ROOT+"/devel/sage/", SAGE_ROOT+"/devel/sage/c_lib", SAGE_ROOT+"/devel/sage/sage/ext", "mtx2.2.4/src/", "pGroupCohomology/c_sources", "pGroupCohomology"]
              ),
    Extension("pGroupCohomology.dickson", sources = ["pGroupCohomology/dickson.pyx"])
    ],
  cmdclass = {'build_ext': build_ext}
)

#######################
## Take care of the data base
tar_base = SAGE_ROOT+'/local/pGroupCohomology/'
from sage.misc.misc import SAGE_DATA
# create data base, if necessary
data_base = SAGE_DATA + '/pGroupCohomology/'
while ('//' in data_base):
    data_base = data_base.replace('//','/')
try:
    os.mkdir(data_base)
except OSError, msg:
    if not ("File exists" in msg):
        raise OSError, msg

# transfer/extract data
os.rename(tar_base+'LICENCE', data_base+'LICENCE')
L = [tar_base+X for X in os.listdir(tar_base) if X.endswith('.tar.gz')]
import tarfile, sys
print 'Extracting data base...'
sys.stdout.flush()
for f in L:
    T = tarfile.open(f)
    T.extractall(path=data_base)
    T.close()
    os.unlink(f)
